from util_process.operators import is_register_operator,is_domain_ONNX
from warnings_thread_safe import warnings 

from util_process.data import get_opset_data,function_data_find,opset_data_find,function_key,opset_key,get_func_values,get_functions_data
from node_connections import check_connections_node,add_connections_to_node_name_lists,get_initializers_names,get_graph_peripherials

from pre_process_module.util_process.util import remove_extra_params
from pre_process_module.repair_types.reapair_value_info import revise_value_infos_graph
from pre_process_module.util_process.util import LINE_WITH_TAB,create_unique_names
from pre_process_module.repair_types.repair_metadata_prop import check_metadata_entries

from pre_process_module.repair_types.repair_annotation import check_annotations
from repair_types.repair_initializer import repair_initializers

MAX_RECURSIVE=1

def __get_correct_params_func_node__(op_name,func_data,import_data,domain=None,overload=None): 
    from pre_process_module.repair_functions.repair_opsets import is_function_contradictory
    
    possible_funcs_name = function_data_find(func_data,name=op_name)
    if len(possible_funcs_name)==0:
        # no existe ninguna funcion con ese nombre
        return False,None,None,None,None
    # hay dos opciones o el dominio no existe o si existe esta incorrecto
    
    if overload==None:
        overload_check=""
    else:
        overload_check=overload
    if domain==None:
        domain_check=""
    else:
        domain_check=domain
    
    possible_funcs_all = function_data_find(func_data,domain=domain_check,name=op_name,overload=overload_check)
    
    # all attributes match -> all correct
    if len(possible_funcs_all)==1:
        return True,possible_funcs_all[0]["domain"],possible_funcs_all[0]["overload"],possible_funcs_all[0]["version"],[{"version":possible_funcs_all[0]["version"]}]
    
    
    if len(possible_funcs_all)>1:
        # same version,domain and overload
        raise RuntimeError("There cannot be two functions with same domain/name/overload at func_data")

    non_contradictory={}

    for fun in possible_funcs_name:
        if not is_function_contradictory(op_name,fun["domain"],import_data):
            non_contradictory[function_key(fun["domain"],fun["name"],fun["version"])]=fun
    if len(non_contradictory)==0:
        #todas las funciones contradicen con el opset conclusion no es una funcion
        return False,None,None,None,None
    
    
        
    
    values=function_data_find(non_contradictory,name=op_name)
    # caso visto anterior pero con la lista filtrada
    if len(values)==1:
        # Estamos seguros que no es contradictorio y hay una solucion
        return True,values[0]["domain"],values[0]["overload"],values[0]["version"]
    if len(values)==0:
        # esta opcion deberia haberse tratado antes
        return False,None,None,None,None
    
    
    func_domains = function_data_find(non_contradictory,domain=domain_check,name=op_name) 
    
    
    unique_domains=get_func_values(non_contradictory,"domain")
    
    if len(func_domains)==0 and (len(unique_domains)>1 or len(unique_domains)<=0):
        # no hay funciones con ese dominio y no es posible ajustar a un dominio concreto
        return False,None,None,None,None
    
    # el dominio existente es incorrecto
    if len(func_domains)==0 and len(unique_domains)==1:
        check,_=is_function_contradictory(name=op_name,domain=unique_domains[0],domains_opset=import_data)
        
        other_domains = opset_data_find(import_data,domain)
        if len(other_domains)>=1:
            #estamos en el posible caso de que existen dominios importados con el dominio incorrecto no es que este mal puesto -> no es una funcion
            return False,None,None,None,None
        if len(other_domains)==0 and check:
            #no hay dominios con este dominio ("erroneo") y el nuevo dominio tampoco es contradictorio
            domain = unique_domains[0]
             
    if len(func_domains)==1:
        return True,func_domains[0]["domain"],func_domains[0]["overload"],func_domains[0]["version"],[{"version":func_domains[0]["version"]}]
    
    #len de func_domains es mayor que 1, hay mas de un posible overload y como ya hemos comparado que el overload vacio sea "",
    #estamos en el caso que hay mas de un overload valido y el overload usado en en la funcion no es ninguno de ellos y no 
    # podemos arreglarlo
    
    final_list = function_data_find(non_contradictory,domain=domain_check,name=op_name,overload=overload_check)
    
    if len(final_list)==1:
        return True,final_list[0]["domain"],final_list[0]["overload"],final_list[0]["version"],[{"version":final_list[0]["version"]}]
    
    # o len(final_list)==0 o len(final_list) es mayor que 1 que ya es raro de por si
    return False,None,None,None,None
        
from onnx.defs import OpSchema

def __get_max_version__(list,key="version"):
    max_version=None
    for item in list:
        version=None
        if isinstance(item,dict):
            version =item[key]
        if isinstance(item,object) and hasattr(item, key):
            version =getattr(item,key)
            
        if max_version==None and version:
            max_version=version
        elif version:
            max_version =max(int(version),int(max_version))
            
    return max_version

# devuelve since_versions que sean menores iguales que version_maximum, de esta forma podemos comprobar si
# el domain opset incluye el operador
def __get_adequate_version__(list_op:list[OpSchema],version_maximum):
    node_version=None
    opset_versions=[]
    deprecated_found=False
    
    list_op.sort(key=lambda x: x.since_version)
    
    for operator in list_op:
        version = operator.since_version
         
        if int(version) <= int(version_maximum) and  (not node_version or int(version)>= int(node_version)):
            if operator.deprecated:
                # si hay una version mayor justo despues (sorted) pero deprecated el operador no esta disponible
                deprecated_found=True
            else:
                # guardamos la version del operador, 
                # en caso de que la version siguiente este deprecated (podemos usar la ultima version disponible)
                # y en caso de que se vuelva a reintroducir en versiones posteriores deprecated_found permite seguir buscando
                node_version=version
                deprecated_found=False
    
    # esta version puede ser escalada en el opset_final (force=False)
    opset_versions=[{"version":version_maximum,"force":False}]
    
    
           
    if deprecated_found:
        # necesitamos ambas versiones la actual y la vieja
        
        if version_maximum>node_version:
            #necesitamos ambas versiones la del nodo y la del opset actual
            opset_versions=[{"version":node_version,"force":True},opset_versions[0]]
            #TODO: improve warning
            warnings.warn("Depracted Operator, actual version is higher")
        else:
            #necesitamos mantener la version del depracated
            opset_versions=[{"version":node_version,"force":True}]
            #TODO: improve warning
            warnings.warn("Depracted Operator, actual version is equal to last functional_version")
        
    return node_version,opset_versions

def __get_opset_version_onnx__(import_data,domain):
    
    
    actual_opsets = opset_data_find(import_data,domain)
   
    
    actual_version=None
    if len(actual_opsets)==1:
        actual_version=actual_opsets[0]["version"]
    elif len(actual_opsets)>1:
        actual_version=__get_max_version__(actual_opsets,key="version")
        
    return actual_version


def __get_correct_params_onnx_node__(op_name,import_data,domain=None):
    
    
    
    list_op:list[OpSchema]=[]
    check,list_op = is_register_operator(op_name)
    require_opset_versions=[]
    
    
    if not check:
        return False,None,None,None
    if len(list_op)==1 and not list_op[0].deprecated:
        # opset_version es la version actual del opset si existe
        
        opset_version=__get_opset_version_onnx__(import_data,list_op[0].domain)
        if opset_version:
            return True,list_op[0].domain,list_op[0].since_version,[{"version":opset_version,"force":False}]
        # como solo hay una entrada no deberian ser operadores depracted
        
    elif len(list_op)==1:
        domain_new = list_op[0].domain
        if domain==None or domain != domain_new:
            domain=domain_new
    
    
    if domain==None:
        domain_check=""
    else:
        domain_check=domain
    
    
    
    check,list_op = is_register_operator(op_name,domain=domain_check)
    
    if not check:
        return False,None,None,None
    if len(list_op)==1 and  not list_op[0].deprecated:
        # como solo hay una entrada no deberian ser operadores depracted
        
        opset_version=__get_opset_version_onnx__(import_data,list_op[0].domain)
        if opset_version:
            return True,list_op[0].domain,list_op[0].since_version,[{"version":opset_version,"force":False}]
    elif len(list_op)==1:
        warnings.warn(f"\nDeprecated operator {list_op[0].name} since version {list_op[0].since_version}")

    
    # si opset_version es None esto implica que hay que añaldir correct_version
    opset_version=__get_opset_version_onnx__(import_data,domain_check)
    
    
    
    correct_version=None
    if opset_version:
        # hay que encontrar en la lista la mejor version para el nodo y la version del opset que se debe comprobar
        # correct_version es la version del nodo (since_version) y new_opset_version es la version que debe existir en los
        # opsets para que sea valido el operador (considera tambien el caso de operadores depracted en versiones anteriores) 
        # poniendo warning -> o dejarlo arreglado o actualizar operador (deberia ser problema del usuario)
        correct_version,require_opset_versions = __get_adequate_version__(list_op,opset_version)
        
    else:
        #no hay ninguna version en el opset para este dominio usamos la mejor version
        correct_version=__get_max_version__(list_op,key="since_version")
        require_opset_versions=[{"version":correct_version,"force":False}]
        
    
    
    # tenemos multiples versiones para el mismo operador, pero como son operadores del dominio onnx, podemos estar seguros cogiendo la 
    # version mayor
    
    if correct_version==None:
        raise RuntimeError("Repairing a ONNX node, There is no version at the opset, revise opsets for missing versions, or execute repair_opsets")
    
    
    return True,domain_check,correct_version,require_opset_versions
        
    
def __get_correct_params_user_domain_node__(op_name,domain,import_data):
    
    list_opset = opset_data_find(import_data,domain)
    
    
    if len(list_opset)==1:
        return True,domain,list_opset[0]["version"]
    max_version=__get_max_version__(list_opset,key="version")
    if len(list_opset)>=1:
        max_version=__get_max_version__(list_opset,key="version")
        if max_version==None:
            #TODO: improve error messege
            raise RuntimeError("Reparing a node, There is no version at the opset, revise opsets for missing versions, or execute repair_opsets")
        return True,domain,max_version
    
    # tenemos multiples versiones para el mismo operador, pero como son operadores del dominio onnx, podemos estar seguros cogiendo la 
    # version mayor
    return False,None,None 

  
  
  
  



def __get_opsets_to_add__(op_name,domain,require_versions,import_data,node_name,func_name,node_pos):
    configs_to_add=[]
    
    
    for require in require_versions:
        
        opset_list = opset_data_find(import_data=import_data,domain=domain,version=require["version"])
        if len(opset_list)>1:
            raise RuntimeError("Opsets in Opsets_Data have two instances with the same key (domain,version)")
    
        if len(opset_list)==0 and "force" in require and require["force"]:
            warnings.warn(f"\nNode {node_name} applaying operator ({op_name}) defined at the graph at the pos {node_pos}. It applies the operator ({op_name}) which is deprecated method in newest versions. \n"+
                          "The last version containing a avaible version of the operator is not write at the opset_import, this version will be added")
            configs_to_add.append({"domain":domain,"version":require["version"],"force":True})
        elif len(opset_list)==0:
            # no esta en la lista de opsets pero luego cuando se revisen solo la version mas grande debe quedar
            
            #warnings.warn(f"\nNode {node_name} applaying operator ({op_name}) defined at function {func_name} at the pos {node_pos}. It applies the operator ({op_name}) but opset_import has its import missing or with a incorrect version")
            configs_to_add.append({"domain":domain,"version":require["version"],"force":False})
            
      
    return configs_to_add




# usar los valores del nodo para determinar si es un operador ONNX,funcion o operador no ONNX 
# y revisar sus atributos de domain,version o overload
def __get_correct_operator_data__(op_name,node,import_data:dict,func_data:dict,func_name="",node_pos=""):
    
    correct=False
    version_correct=None
    domain_correct=None
    overload_correct=None
    configs_to_add=None
    is_onnx=False
    is_custom=False
  
    
    is_function=False
    domain=None
    
    overload=None
    
    node_name = ""
    if "name" in node:
        node_name=f"({node["name"]})"
    
    
    
    if "domain" in node:
        domain =node["domain"]
    if "overload" in node:
        overload =node["overload"]
    
    
    is_function,domain_correct,overload_correct,version_correct,require_opset_versions = __get_correct_params_func_node__(op_name,func_data,import_data,domain,overload)
    
    
    if is_function and op_name==func_name:
        raise SyntaxError(f"Node {node_name} defined at the function {func_name} at the pos {node_pos}. It calls itself. Recursive calls are not allowed")
    
    
    
    if not is_function and overload:
        del node["overload"]

    
    if not is_function:
        # ahora o es un operador onnx o un operator set, custom que no sea a añadido a la lista de operators del motor onnx con is register
        # y solo aparece en la lista de dominios
        
        
        is_onnx,domain_correct,version_correct,require_opset_versions=__get_correct_params_onnx_node__(op_name,import_data,domain)
        
    
    
    if not is_onnx and not is_function and domain!=None:
        
        is_custom,domain_correct,version_correct = __get_correct_params_user_domain_node__(op_name,domain,import_data)
        
       
        
        


    
    
    if not is_onnx and not is_function and not is_custom:
        if domain==None and func_name!="": 
            raise SyntaxError(f"Node {node_name} defined at the graph at the pos {node_pos}.It applies the operator with name ({op_name}) but the operator has not domain and it is not related to any known operator or defined opset. Fix the Node")
        elif domain==None:
            raise SyntaxError(f"Node {node_name} defined at the function {func_name} at the pos {node_pos}. It applies the operator with name ({op_name}) but the operator is not related to any known operator or defined opset. Fix the Node")
    
    #check_func_node(op_name,node,versions,func_data)
    

        
    if is_onnx or is_function:
        
        configs_to_add=__get_opsets_to_add__(op_name,domain_correct,require_opset_versions,import_data,node_name,func_name,node_pos)  
        
    
    if overload_correct!=None:
        result = {"opType":op_name,"domain":domain_correct,"version":version_correct,"overload":overload_correct} 
    else:       
        result={"opType":op_name,"domain":domain_correct,"version":version_correct}
        
    correct=is_function or is_onnx or is_custom

    return correct,result,configs_to_add,(is_function,is_onnx,is_custom)













def __repair_default_node__(node,correct_data,func_name="",node_pos=""):
    
         
    #No esta domain pero la version es igual a la version onnx (onnx_version)
    node_name = ""
    new_overload=None
    old_domain=None
    old_overload=None
    if "name" in node:
        node_name=f"({node["name"]})"
    
    if "domain" in node:
        old_domain=node["domain"]
    
    if "overload" in node:
        old_overload=node["overload"]  
    if  "overload" in correct_data:
        new_overload=correct_data["overload"]
    new_domain=correct_data["domain"]
    
    
    node["version"]=str(correct_data["version"])
      
      
    if old_domain!=None and new_domain!=old_domain :
        if func_name=="":
            warnings.warn(f"\nNode {node_name} defined at the graph at the pos {node_pos}.Its domain ({old_domain}) was changed for correct domain ({new_domain}) ")
        else:
            warnings.warn(f"\nNode {node_name} defined at the function {func_name} at the pos {node_pos}.Its domain ({old_domain}) was changed for correct domain ({new_domain}) ")
    node["domain"]=new_domain    
     
    if new_overload!=None and old_overload!=None and new_overload!=old_overload:
        if func_name=="":
            warnings.warn(f"\nNode {node_name} defined at the graph at the pos {node_pos}.Its function overload ({old_overload}) was changed for correct domain ({new_overload}) ")
        else:
            warnings.warn(f"\nNode {node_name} defined at the function {func_name} at the pos {node_pos}.Its function overload ({old_overload}) was changed for correct overload ({new_overload}) ")
    if new_overload!=None: 
        node["overload"]=new_overload
    

# prerequisite: using repair_opset_and_funcs with data                


def __combine_config_to_add__(all_configs_add:dict,configs_add_local):
    
    for config in configs_add_local:
        key=opset_key(config["domain"],config["version"])
       
        if key not in all_configs_add.keys() and config["force"] :
            #estos son obligatorios
            all_configs_add[key]=config
        
        configs_domain = opset_data_find(all_configs_add,domain=config["domain"])
        if len(configs_domain)==0:
             all_configs_add[key]=config
        if len(configs_domain)>=1:
            max_version = __get_max_version__(configs_domain,"version")
            if max_version<config["version"]:
                all_configs_add= __delete_config_to_add__(all_configs_add,configs_domain)
                all_configs_add[key]=config
                
def __delete_config_to_add__(all_configs_add:dict,list_config:list[dict]):
    for config in list_config:
        key=opset_key(config["domain"],config["version"])
        if not config["force"]:
            del all_configs_add[key]
    return all_configs_add


  
def __repair_node_name__(node,all_names):
    if "name" in node:
        base_name=node["name"]
    else:
        base_name="unnamed"
    
    new_name =create_unique_names(all_names,base_name)    
    all_names.append(new_name)
    
    return new_name
    
def repair_nodes(data):
    from pre_process_module.repair_functions.repair_opsets import repair_opset_and_funcs

    #main graph
    nodes = data["graph"]["node"]
    
    imports = data["opsetImport"]
    func_data=get_functions_data(data)
    
    
    all_initlizer_names=dict()
    get_initializers_names(data["graph"]["initializer"],all_initlizer_names)
    get_initializers_names(data["graph"]["sparseInitializer"],all_initlizer_names,is_sparse=True)
    
    peripherials = get_graph_peripherials(data["graph"])
    sub_graph_metadata={"global_data":data,"graph_names":[data["graph"]["name"]]}
    correct =__repair_nodes_local__(nodes,imports,func_data,initializers=all_initlizer_names,peripherials=peripherials,graph_name=data["graph"]["name"],sub_graph_metadata=sub_graph_metadata)
    iterations=0
    
    while not correct:
        iterations=iterations+1
        repair_opset_and_funcs(data)
        
        correct =__repair_nodes_local__(nodes,imports,func_data,initializers=all_initlizer_names,peripherials=peripherials,graph_name=data["graph"]["name"],sub_graph_metadata=sub_graph_metadata)
        if not correct and iterations==MAX_RECURSIVE:
            raise RuntimeError("Nodes at Graph cannot be repaired")
    
    for function in data["functions"]:
        nodes = function["node"]
        imports = function["opsetImport"]
        
        peripherials = get_graph_peripherials(function)
        
        
        correct =__repair_nodes_local__(nodes,imports,func_data,func_name=function["name"],initializers=all_initlizer_names,peripherials=peripherials,global_func=function)
        iterations=0
        while not correct:
            repair_opset_and_funcs(data)
            correct =__repair_nodes_local__(nodes,imports,func_data,func_name=function["name"],initializers=all_initlizer_names,peripherials=peripherials,iterations=iterations,global_func=function)
            if not correct and iterations==MAX_RECURSIVE:
                raise RuntimeError(f"Nodes at Function ({function["name"]}) cannot be repaired")                              
    
def __repair_nodes_local__(nodes,imports,func_data,func_name="",graph_name="",initializers=[],peripherials=None,iterations=0,up_element_id="",sub_graph_metadata=None,global_func="",start_error_messege="\n"):
    from pre_process_module.repair_types.repair_attributes import check_attributes
    from pre_process_module.repair_functions.check_functions import get_func_element_id
    
    imports_data = get_opset_data(imports) 

    
    global_data = None
    if sub_graph_metadata and "up_metadata" not in sub_graph_metadata:
        up_metadata=dict()
    else:
        up_metadata = sub_graph_metadata["up_metadata"]
        
    if sub_graph_metadata and "global_data" in sub_graph_metadata:
        global_data = sub_graph_metadata["global_data"]
        
    
    all_config_add=dict()
    types_operator_data=[]
    correct_all=True  

    all_names=[]
    input_names=dict()
    output_names=dict()
    
    
    element_id=""
    if func_name!="":
        element_id=f"Func-{get_func_element_id(global_func)}"
        up_metadata[func_name]={"inputs":input_names,"outputs":output_names,"initializers":initializers,"graph_peripherials":peripherials}
    
    if graph_name!="":
        element_id=f"Graph-{graph_name}"
        up_metadata[graph_name]={"inputs":input_names,"outputs":output_names,"initializers":initializers,"graph_peripherials":peripherials}
        
    if up_element_id!="":
        element_id=f"{up_element_id}-{element_id}"
    
    
    
    for idx,node in enumerate(nodes):
        new_name = __repair_node_name__(node,all_names)
        node["name"]=new_name
        
        if not __check_node_fields__(node):
            if func_name!="":
                warnings.warn(f"{start_error_messege}Node number {idx} of function ({func_name}) doesnt follow correct format: It is not a dict. Deleting node ")
            else:
                warnings.warn(f"{start_error_messege}Node number {idx} of the graph ({graph_name}) doesnt follow correct format: It is not a dict. Deleting node ")
            continue
        
        if iterations==0:
            add_connections_to_node_name_lists(node,output_names,input_names,element_id)
        
        if "opType" in node:
            check,correct_data,configs_to_add,type_op=__get_correct_operator_data__(node["opType"],node,imports_data,func_data,func_name=func_name,node_pos=idx)
            
            if not check:
                correct_all=False
                __clean_node__(node,False,start_error_messege=start_error_messege)
                types_operator_data.append(None)
                continue
            
            
            if configs_to_add:
                # version en opset invalida para el operador usado, guardando datos para reparar
                __combine_config_to_add__(all_config_add,configs_to_add)
                
                     
            
            if "overload" in correct_data.keys():
                repairs = {"opType":node["opType"],"domain":correct_data["domain"],"overload":correct_data["overload"],"version":correct_data["version"]}
                __repair_default_node__(node,repairs,func_name,idx)
                
            else:
                repairs = {"opType":node["opType"],"domain":correct_data["domain"],"version":correct_data["version"]}
                __repair_default_node__(node,repairs,func_name,idx)
            types_operator_data.append(type_op)
        else:
            types_operator_data.append(None)
            
        __clean_node__(node,True,start_error_messege)
                
    
    
    
    if len(all_config_add)>0:
        for config in all_config_add.values():
            imports.append({"domain":config["domain"],"version":config["version"]})            
    if len(all_config_add)>0:
        return False
    
    
    
      
    
    
    
    for idx,node in enumerate(nodes):
        
        node["element_id"]=element_id
        func_data_conn = None
        if global_func!="":
            func_data_conn=global_func
        if "opType" in node:
            check_connections_node(node,conn_data=up_metadata,element_id=element_id,
                                   is_onnx=types_operator_data[idx]!= None and types_operator_data[idx][1],
                                   func_data=func_data_conn,
                                   start_error_messege =f"{start_error_messege}nConnections of node of graph {graph_name} at pos {idx} with name ({node["name"]}) have a wrong format {LINE_WITH_TAB}")
            if types_operator_data[idx] and types_operator_data[idx][0]:
                node["func_id"]=build_applies_func_id(node)
                
                get_func_element_id()
                
            if types_operator_data[idx] and types_operator_data[idx][1] :
                node["operator_id"]=build_applies_operator_id(node)
                
            if types_operator_data[idx] and types_operator_data[idx][2]:
                node["custom_id"]=build_applies_operator_id(node)
            
        attributes=[]
        if "attribute" in node:
            attributes=node["attribute"]
        corret_all_att=True
        
        correct_att,corret_all_att=check_attributes(attributes,node,types_operator_data[idx],func_data,element_id=f"{element_id}-Node-{node["name"]}",
                                   subgraph_data={"global_data":global_data,"up_metadata":up_metadata}
                                   ,start_error_messege=f"{start_error_messege} Attributes of node of graph {graph_name} at pos {idx} with name ({node["name"]}) have a wrong format {LINE_WITH_TAB}")
        
        if "metadataProps" in node:
            correct_meta,_ =check_metadata_entries(node["metadataProps"],start_error_messege=f"{start_error_messege}The nodo of graph {graph_name} at pos {idx} with name ({node["name"]}) has a wrong metadata_prop param {LINE_WITH_TAB}",keys=[],element_id=element_id)

            if not correct_meta:
                node["metadataProps"]=[]
        
        
        
        if not corret_all_att or not correct_att:
            correct_all=False
            
            if func_name!="":
                warnings.warn(f"{start_error_messege}Node number {idx} of function ({func_name}) has invalid attribute param.")
            else:
                 warnings.warn(f"{start_error_messege}Node number {idx} of the graph ({graph_name}) has invalid attribute param.")
    
         
               
    if not correct_all:
        raise SyntaxError("While repairng the nodes some error where find on some of the nodes: Operators are invalid or connections are wrong or attributes are invalid. Revise and fix the model")               
    
    return True

def build_applies_operator_id(node):
    
    domain_str=""
    if node["domain"]!="":
        domain_str=f"-domain-{node["domain"]}"
    
    return f"name-{node["opType"]}{domain_str}-v{node["version"]}"
    
def build_applies_func_id(node):
    overload_str=""
    if node["overload"]!="":
        overload_str=f"-overload-{node["overload"]}"
    
    return f"Func-{node["opType"]}-domain-{node["domain"]}{overload_str}"

def __check_sub_graph_fields__(graph,start_error_messege="",graph_names:list[str]=[]):
    if "node" not in graph or not isinstance(graph["node"],list):
            graph["node"]=[]
    if "initializer" not in graph or not isinstance(graph["initializer"],list):
            graph["initializer"]=[]
            
    if "sparseInitializer" not in graph or not isinstance(graph["sparseInitializer"],list):
            graph["sparseInitializer"]=[]
    
    if "name" in graph :
        base_name =graph["name"]
    else:
        base_name="unnamed_graph"   
        
    graph["name"] =create_unique_names(graph_names,base_name)
        
            
    return True



def check_subgraph(subgraph,subgraph_data,start_error_messege="",element_id=""):
    subgraph["element_id"]=element_id
    subgraph["is_subgraph"]=""
    global_data = subgraph_data["global_data"]
    up_metadata = subgraph_data["up_metadata"]
  
  
    __check_sub_graph_fields__(subgraph,start_error_messege=f"{start_error_messege}",graph_names=up_metadata.keys())

    graph_element_id = f"{element_id}-Graph-{subgraph["name"]}"
    
    repair_initializers(subgraph,start_error_messege=f"{start_error_messege} The subgraph defined as an attribute value has a wrong format {LINE_WITH_TAB}")
    
    revise_value_infos_graph(subgraph,element_id=graph_element_id,start_error_messege=f"{start_error_messege} The subgraph defined as an attribute value has a wrong format {LINE_WITH_TAB}")
    
    
    
    imports = global_data["opsetImport"]
    func_data=get_functions_data(global_data)
    
    all_initlizer_names=dict()
    get_initializers_names(subgraph["initializer"],all_initlizer_names,start_error_messege=start_error_messege)
    get_initializers_names(subgraph["sparseInitializer"],all_initlizer_names,is_sparse=True,start_error_messege=start_error_messege)
    peripherials = get_graph_peripherials(subgraph["node"])
    
    if "metadataProps" in subgraph:
        correct_meta,_ =check_metadata_entries(subgraph["metadataProps"],start_error_messege=f"{start_error_messege}The graph {subgraph["name"]} has a wrong metadata_prop param {LINE_WITH_TAB}",keys=[],element_id=element_id)

        if not correct_meta:
            subgraph["metadataProps"]=[]

    __repair_nodes_local__(subgraph["node"],imports,func_data,graph_name=subgraph["name"],peripherials=peripherials,initializers=all_initlizer_names,up_element_id=element_id,sub_graph_metadata=subgraph_data,
                           start_error_messege=f"{start_error_messege}")
    
    check_annotations(subgraph,element_id,start_error_messege = f"{start_error_messege} The quantizationAnnotation of graph ({subgraph["name"]}) has a wrong format {LINE_WITH_TAB}" )
    
    remove_extra_params(subgraph,["input","output","node","name","initializer","sparseInitializer","valueInfo","element_id","metadataProps","docString","quantizationAnnotation","is_subgraph"],
                        f"The subgraph with name ({subgraph["name"]})")
    return True






def __clean_node__(node,correct,start_error_messege=""):
    keep_dims=["opType","domain","overload","version","name","input","output","attribute","docString","metadataProps","is_custom"]
    keep_dims_wrong=["name","input","output","docString","metadataProps"]
    if correct:
        remove_extra_params(node,keep_dims,start_error_messege=f"The node with name ({node["name"]})")
    else:
        remove_extra_params(node,keep_dims_wrong,start_error_messege=f"The node with name ({node["name"]})")
   
def __check_node_fields__(node):
    if not isinstance(node,dict):
        return False
    
            
    return True





from warnings_thread_safe import warnings 
from onnx.defs import  OpSchema
from pre_process_module.util_process.util import DOUBLE_LINE_WITH_TAB
from pre_process_module.util_process.operators import find_operators

def add_connections_to_namelist(connections:list,name_list:dict,node_id,start_error_messege=""):
    compose_warning=f"{start_error_messege}"
    warning_present=False
    for idx,conn in enumerate(connections):
        if isinstance(conn,str) and conn not in name_list.keys():
            name_list[conn]=f"{node_id}-idx{idx}"
        elif not isinstance(conn,str):
            warning_present=True
            compose_warning= f"{compose_warning} {DOUBLE_LINE_WITH_TAB} Connection at pos {idx} is not a string -> Being Skipped"
    if warning_present:
        warnings.warn(compose_warning)        
            
def get_specific_peripherials(peripherials:list,start_error_messege=""):
    peripherial_name_list=dict()
    compose_warning=f"{start_error_messege}"
    warning_present=False
    for idx,peripherial in enumerate(peripherials):
        if isinstance(peripherial,dict) and "name" in peripherial.keys() and peripherial["name"] not in peripherial_name_list.keys():
            peripherial_name_list[peripherial["name"]]=peripherial["element_id"]
        elif isinstance(peripherial,str):
            peripherial_name_list[peripherial]=""
        else:
            warning_present=True
            compose_warning= f"{compose_warning} Peripherial at pos {idx} is not a valid entry -> Being Skipped"
    if warning_present:
        warnings.warn(compose_warning)
    
    return peripherial_name_list   
     
def get_initializers_names(initializers:list,name_list:dict,start_error_messege="",is_sparse=False):
    compose_warning=f"{start_error_messege}"
    warning_present=False
    for idx,initliazer in enumerate(initializers):
        
        if isinstance(initliazer,dict) and "name" in initliazer.keys() and initliazer["name"] not in name_list.keys():
            name_list[initliazer["name"]]={"id":initliazer["tensor_id"],"is_sparse":is_sparse}
        else:
            warning_present=True
            compose_warning= f"{compose_warning} Initializer at pos {idx} is not a valid entry -> Being Skipped"  
    if warning_present:
        warnings.warn(compose_warning)        
            
            
def get_graph_peripherials(graph,start_error_messege=""):
    inputs=dict()
    outputs=dict()
    if "input" in graph:
        inputs = get_specific_peripherials(graph["input"],
            start_error_messege=f"{start_error_messege} Inputs of graph {graph["name"]} have a wrong format") 
    if "output" in graph:
        outputs = get_specific_peripherials(graph["output"],
            start_error_messege=f"{start_error_messege} Outputs of graph {graph["name"]} have a wrong format")   
 
    return {"inputs":inputs,"outputs":outputs}      
            
def add_connections_to_node_name_lists(node,output_names:dict,input_names:dict,node_id="",start_error_messege=""):
    if "output" in node:
        add_connections_to_namelist(node["output"],output_names,f"{node_id}-Node-{node["name"]}-outputs",
                                    start_error_messege=f"{start_error_messege} Output connections have a wrong format")
    if "input" in node:
        add_connections_to_namelist(node["input"],input_names,f"{node_id}-Node-{node["name"]}-inputs",
                                    start_error_messege=f"{start_error_messege} Input connections have a wrong format")
    

    # lista mas de un elemento
    
def check_for_variadic(peripherial):
    variadic_found=False
    for idx,item in enumerate(peripherial):
        if item.option.name == "Variadic":
            variadic_found=True
        if variadic_found and item.option.name != "Variadic":
            return False,idx,variadic_found
    return True,idx,variadic_found
    
def check_peripherials_onnx(op_schema:OpSchema,is_output=False,connections=[],start_error_messege=""):
    if not op_schema:
        warnings.warn(f"{start_error_messege} Onnx operator metadata is empty")
        return False,False,None
    peripherial =None
    
    if is_output:
        peripherial=op_schema.outputs
    else:
        peripherial=op_schema.inputs
        
    is_variadic_correct,idx,is_variadic = check_for_variadic(peripherial)
    
    
    if not is_variadic_correct:
        warnings.warn(f"{start_error_messege} Onnx operator {op_schema.name} has a variadic variable (dynamic number of inputs/outputs) but it has (Single,Optional) after the variadic input at input pos {idx} ")
        return False,is_variadic,peripherial
    

    
    if not is_variadic and len(peripherial)<len(connections):
        warnings.warn(f"{start_error_messege} Onnx operator {op_schema.name} is not variadic or accepts more than {len(peripherial)} inputs ")
        return False,is_variadic,peripherial
   
    return True,is_variadic,peripherial
      
    
def check_optional_variables(op_schema,peripherial_onnx,idx,start_error_messege=""):

    variable= peripherial_onnx[idx]
    if variable.option.name!="Optional":
        return False,f"{start_error_messege} is empty but \n\t\t\t Onnx operator {op_schema.name} doenst allow the input with pos {idx} to be optional"
    return True,""

def get_new_connection_list(connections,connections_names:dict,node_name,is_output,initializers=[],inputs_graph_names=[],output_graphs_names=[],start_error_messege="",
                            op_schema=None,check_peripherials=False,func_data=None):
    from pre_process_module.repair_functions.check_functions import get_func_element_id
    
    new_connections=[]
    
    compose_warning = f"{start_error_messege}"
    all_correct = True
  
    
    if check_peripherials:
        all_correct,is_variadic,peripherial = check_peripherials_onnx(op_schema,is_output,connections,start_error_messege=start_error_messege)
    if not all_correct:
        warnings.warn(f"{start_error_messege} Connections have a wrong format respecting the operator, check warnings")
    
    
    for idx,connc in enumerate(connections):
        
        if not isinstance(connc,str):
            continue
        
          
        
        if connc!=None and connc in connections_names.keys() or connc in inputs_graph_names.keys() or connc in output_graphs_names.keys() or connc in initializers.keys():
            
            
            
            connection = {"name":connc,"element_id":node_name,"is_empty":False}
            if func_data:
                connection["func_id"] = get_func_element_id(func_data)
            if connc in connections_names.keys(): 
                connection["conc_id"]=connections_names[connc] 
            if connc in inputs_graph_names.keys():
                connection["g_input_id"]=inputs_graph_names[connc]
            if connc in output_graphs_names.keys():
                connection["g_output_id"]=output_graphs_names[connc] 
            if connc in initializers.keys():
                if initializers[connc]["is_sparse"]:
                    connection["sparse_id"]=initializers[connc]["id"]
                else:
                    connection["init_id"]=initializers[connc]["id"] 
                
                
            new_connections.append(connection)
        elif connc!="" and connc!=None:
            
            
            if not is_output:
                compose_warning = f"{compose_warning} {DOUBLE_LINE_WITH_TAB} Node connection ({connc}) of node {node_name} doenst exist "
                all_correct=False
        else:
            correct=True
            if check_peripherials and not is_variadic :
                correct,error = check_optional_variables(op_schema,peripherial,idx,start_error_messege=f"{compose_warning} Node connection ({connc}) of node {node_name}")

            if correct:    
                new_connections.append({"name":"","element_id":node_name,"is_empty":True})
            else:
                all_correct=False
                compose_warning=error       
        
    if not all_correct:
       warnings.warn(compose_warning)    
       
  
    if not all_correct:
        raise SyntaxError(f"{start_error_messege} Connections have a wrong format that cannot be fix, check warnings")
        


    for idx,conn in enumerate(new_connections):
        conn["index"]=idx
        if idx < len(new_connections)-1:
            conn["next_index"]=idx+1
        if idx == len(new_connections)-1:
            conn["last_index"]=""
        
    return new_connections       


def check_connections_node(node,conn_data:dict,element_id="",start_error_messege="",is_onnx=False,func_data=None):

    
    op_schema=None
    if is_onnx:
        op_schema = find_operators(node["opType"],node["domain"],node["version"])
        if len(op_schema)>0:
            op_schema=op_schema[0]
        else:
            raise RuntimeError(f"{start_error_messege} Operator {node["opType"]}-{node["domain"]}-{node["version"]} doenst exist")
   
    input_names = dict()
    output_names = dict()
    initializers = dict()
    graph_peripherials = {"inputs":dict(),"outputs":dict()}
    

    for item in conn_data.values():
        input_names = input_names | item["inputs"]
        output_names = output_names | item["outputs"]
        initializers = initializers |item["initializers"]
        graph_peripherials["inputs"]=graph_peripherials["inputs"] | item["inputs"]
        graph_peripherials["outputs"]=graph_peripherials["outputs"] | item["outputs"]
   
    
    node_name= f"{element_id}-Node-{node["name"]}"
        
    if "input" in node:
        inputs=get_new_connection_list(node["input"],output_names,node_name=node_name,is_output=False,initializers=initializers,
                                            inputs_graph_names=graph_peripherials["inputs"],
                                            output_graphs_names=dict(),
                                            start_error_messege=f"{start_error_messege} Inputs have a wrong format",
                                            func_data=func_data,
                                            check_peripherials=is_onnx,op_schema=op_schema)
        del node["input"]
        node["input_node"]=inputs
        
    if "output" in node:
        outputs=get_new_connection_list(node["output"],input_names,node_name=node_name,is_output=True,initializers=initializers,
                                            inputs_graph_names=dict(),
                                            output_graphs_names=graph_peripherials["outputs"],
                                            start_error_messege= f"{start_error_messege} Outputs have a wrong format",
                                            func_data=func_data,
                                            check_peripherials=is_onnx,op_schema=op_schema)
        del node["output"]
        node["output_node"]=outputs
        
    




from warnings_thread_safe import warnings 
from pre_process_module.util_process.util import remove_extra_params,LINE_WITH_TAB
from repair_tensor import check_tensor_simple,check_tensor_data_type,DataType,get_label_data_type



def check_type_proto(type,start_error_messege:str="",element_id=""):
    if not isinstance(type,dict):
        warnings.warn(f"\n{start_error_messege} Type is not a dict")
        return False
    
    if "denotation" in type and not isinstance(type["denotation"],str):
        warnings.warn(f"\n{start_error_messege} \"denotation\" param is not a string -> Deleting \"denotation\" param")
        
    
    
    if "tensorType" in type:
        check = check_tensor_simple(type["tensorType"],tensor_id=element_id,
            start_error_messege=f"{start_error_messege} The Tensor Value Info has a wrong format {LINE_WITH_TAB}")
        if check:
            remove_extra_params(type,keep_params=["tensorType","denotation"],start_error_messege=f"{start_error_messege} The Tensor Value Info ")
            type["element_id"]=element_id
            type["type"]="tensor"
            return True
        else:
            return False
    if "sparseTensorType" in type:
        check = check_tensor_simple(type,tensor_id=element_id,
            start_error_messege=f"{start_error_messege} The Sparse Tensor Value Info has a wrong format {LINE_WITH_TAB}")
        
        if check:
            remove_extra_params(type,keep_params=["sparseTensorType","denotation"],start_error_messege=f"{start_error_messege} The Sparse Tensor Value Info ")
            type["element_id"]=element_id
            type["type"]="sparse"
            return True
        else:
            return False
    if "sequenceType" in type:
        check = check_sequence_type(type["sequenceType"],f"{start_error_messege} The Sequence Value Info ")
        if check:
            remove_extra_params(type,keep_params=["sequenceType","denotation"],start_error_messege=f"{start_error_messege} The Sequence Value Info ")
            type["element_id"]=element_id
            type["type"]="seq"
            return True
        else:
            return False
    if "mapType" in type:
        check = check_map_type(type["mapType"],f"{start_error_messege} The Map Value Info ")
        if check:
            remove_extra_params(type,keep_params=["mapType","denotation"],start_error_messege=f"{start_error_messege} The Map Value Info ")
            type["element_id"]=element_id
            type["type"]="map"
            return True
        else:
            return False
    if "optional" in type:
        check = check_optional_type(type["optional"],f"{start_error_messege} The Optional Value Info ")
        if check:
            remove_extra_params(type,keep_params=["optional","denotation"],start_error_messege=f"{start_error_messege} The Optional Value Info ")
            type["element_id"]=element_id
            type["type"]="opt"
            return True
        else:
            return False
    
    
    warnings.warn(f"\n{start_error_messege} Type is not a valid Type (tensorType,sparseTensorType,sequenceType,mapType,optionalType)")
    
    return False


def check_sequence_type(value,start_error_messege:str="",element_id=""):
    if not isinstance(value,dict):
        warnings.warn(f"{start_error_messege} Is not a Dict" )
        return False
    if "elem_type" not in value:
        warnings.warn(f"{start_error_messege} \"elem_type\" param is missing" )
        return False
    else:
        if not check_type_proto(value["elem_type"],f"{start_error_messege} \"elem_type\" param has a wrong format {LINE_WITH_TAB}",element_id=f"{element_id}-Seq"):
            return False
    remove_extra_params(value,keep_params=["elem_type"],start_error_messege=start_error_messege)
    return True

def check_map_type(value,start_error_messege:str="",element_id=""):
    if not isinstance(value,dict):
        warnings.warn(f"{start_error_messege} Is not a Dict" )
        return False
    
    if "key_type" not in value:
        warnings.warn(f"{start_error_messege} \"value_type\" param is missing" )
        return False
    
    
    if "value_type" not in value:
        warnings.warn(f"{start_error_messege} \"value_type\" param is missing" )
        return False   
      
    if not check_type_proto(value["value_type"],f"{start_error_messege} \"value_type\" param has a wrong format {LINE_WITH_TAB}",element_id=f"{element_id}-Map"):
        return False
    
    try:
        int(value["key_type"])
    except Exception:
        warnings.warn(f"{start_error_messege} \"key_type\" must be a int value referencing a Tensor.DataType" )
        return False
    enum = check_tensor_data_type(value["key_type"])
    if enum not in [DataType.INT8,DataType.INT16,DataType.INT32,DataType.INT64,
                    DataType.UINT8,DataType.UINT16,DataType.UINT32,DataType.UINT64,DataType.STRING]:
        warnings.warn(f"{start_error_messege} \"key_type\" can only be of DataType ([U]INT{{8|16|32|64}}) or STRING" )
        return False
    label = get_label_data_type(enum)
    value["key_type"]={"prefLabel":enum.name,"altLabel":enum.value,"rdfLabel":label}
    remove_extra_params(value,keep_params=["key_type","value_type"],start_error_messege=start_error_messege)
    return True

def check_optional_type(value,start_error_messege:str="",element_id=""):
    if not isinstance(value,dict):
        warnings.warn(f"{start_error_messege} Is not a Dict" )
        return False
    if not check_type_proto(value["elem_type"],f"{start_error_messege} \"elem_type\" param has a wrong format {LINE_WITH_TAB}",element_id=f"{element_id}-Opt"):
            return False
    remove_extra_params(value,keep_params=["elem_type"],start_error_messege=start_error_messege)
    return True



from warnings_thread_safe import warnings 
from repair_metadata_prop import check_metadata_entries
from repair_type_proto import check_type_proto
from pre_process_module.util_process.util import remove_extra_params,LINE_WITH_TAB

def check_value_infos(value_infos,start_error_messege:str="",force_type=False,element_id=""):
    
    
    if not isinstance(value_infos,list):
        warnings.warn(f"\n{start_error_messege} valueInfo is not a list -> valueInfo being deleted")
        return False
    correct=True
    names=[]
    for idx,entry in enumerate(value_infos):
        correct = check_value_info_entry(entry,names,start_error_messege=f"{start_error_messege} valueInfo entry at pos {idx}{LINE_WITH_TAB}",force_type=force_type,element_id=element_id)
        if not correct:
            del value_infos[idx]
            
    return len(value_infos)>0

def check_value_info_entry(value_info,names,start_error_messege:str="",force_type=False,element_id=""):
    if "name" not in value_info:
        warnings.warn(f"\n{start_error_messege} has no param name -> valueInfo entry being deleted")
        return False
    if value_info["name"] in names:
        warnings.warn(f"\n{start_error_messege} its param name is not unique (Duplicate names not allowed) -> duplicate valueInfo entry being deleted")
        return False
    if "docString" in value_info and not isinstance(value_info["docString"],str):
        warnings.warn(f"\n{start_error_messege} The \"docString\" is not a string -> Removing \"docString\"")
        value_info["docString"]=""
    if "metadataProps" in value_info:
        keys=[]
        correct,_ = check_metadata_entries(value_info["metadataProps"],keys,element_id=element_id,start_error_messege=f"{start_error_messege} metadataProps param has a wrong format")
        
        if not correct:
            value_info["metadataProps"]=[]
        
    if "type" in value_info:
        correct= check_type_proto(value_info["type"],start_error_messege=f"{start_error_messege} type param has a wrong format -> valueInfo entry being deleted{LINE_WITH_TAB}",
                                  element_id=f"{element_id}-{value_info["name"]}")
        if not correct:
            return False
    elif force_type:
        warnings.warn(f"\n{start_error_messege} type param is missing -> valueInfo entry being deleted")
        return False
    remove_extra_params(value_info,["name","docString","metadataProps","type","is_input","is_output","is_value"],start_error_messege)
    value_info["element_id"]=f"{element_id}-{value_info["name"]}"
    return True



def revise_value_infos_graph(graph,element_id="",start_error_messege=""):
    if element_id!="":
        element_id=f"{element_id}-"
    
    if "valueInfo" in graph:
        for item in graph["valueInfo"]:
            item["is_value"]=""
        _ = check_value_infos(graph["valueInfo"],start_error_messege=f"{start_error_messege} The valueInfo param defined at the graph has a wrong format{LINE_WITH_TAB}",element_id=f"{element_id}Graph-{graph["name"]}-valuesInfo")
    
    if "input" in graph:
        for item in graph["input"]:
            item["is_input"]=""
        _ = check_value_infos(graph["input"],start_error_messege=f"{start_error_messege} The input param defined at the graph has a wrong format{LINE_WITH_TAB}",force_type=True,element_id=f"{element_id}Graph-{graph["name"]}-inputs")
        
    if "output" in graph:
        for item in graph["output"]:
            item["is_output"]=""
        _ = check_value_infos(graph["output"],start_error_messege=f"{start_error_messege} The output param defined at the graph has a wrong format{LINE_WITH_TAB}",force_type=True,element_id=f"{element_id}Graph-{graph["name"]}-outputs")


from repair_tensor import check_tensor_complex,check_sparse_complex
from warnings_thread_safe import warnings 
from pre_process_module.util_process.operators import find_operators
from pre_process_module.util_process.data import function_data_find
from repair_type_proto import check_type_proto
from pre_process_module.util_process.util import remove_extra_params,validate_enum,LINE_WITH_TAB
from pre_process_module.repair_nodes.repair_nodes_base import check_subgraph



from enum import Enum
from warnings_thread_safe import warnings 
from onnx.defs import  OpSchema
class AttrType(Enum):
    UNDEFINED = 0
    FLOAT = 1
    INT = 2
    STRING = 3
    TENSOR = 4
    GRAPH = 5
    SPARSE_TENSOR = 11
    TYPE_PROTO = 13

    FLOATS = 6
    INTS = 7
    STRINGS = 8
    TENSORS = 9
    GRAPHS = 10
    SPARSE_TENSORS = 12
    TYPE_PROTOS = 14
    
TYPED_LISTS=[AttrType.FLOATS,AttrType.INTS,AttrType.STRINGS,AttrType.GRAPHS,AttrType.SPARSE_TENSORS,AttrType.TYPE_PROTOS]
INT_TYPES = [AttrType.INT,AttrType.INTS]
STRING_TYPES = [AttrType.STRING,AttrType.STRINGS]
FLOAT_TYPES = [AttrType.FLOAT,AttrType.FLOATS]
TENSOR_TYPES = [AttrType.TENSOR,AttrType.TENSORS]
SPARSE_TYPES = [AttrType.SPARSE_TENSOR,AttrType.SPARSE_TENSORS]
GRAPH_TYPES = [AttrType.GRAPH,AttrType.GRAPHS]
TYPE_PROTO_TYPES = [AttrType.TYPE_PROTO,AttrType.TYPE_PROTOS]

MAPP_ENUM_LABEL={AttrType.FLOAT:"f",AttrType.FLOATS:"floats",
                 AttrType.INT:"i",AttrType.INTS:"ints",
                 AttrType.STRING:"s",AttrType.STRINGS:"strings",
                 AttrType.TENSOR:"t",AttrType.TENSORS:"tensors",
                 AttrType.GRAPH:"g",AttrType.GRAPHS:"graphs",
                 AttrType.SPARSE_TENSOR:"sparse_tensor",AttrType.SPARSE_TENSORS:"sparse_tensors",
                 AttrType.TYPE_PROTO:"tp",AttrType.TYPE_PROTOS:"type_protos"}



def check_att_enum(data_type_value):
   enum = validate_enum(data_type_value,AttrType)
   
   return enum

def check_onnx_op_attributes(enum_type,attribute,op_schema:OpSchema,start_error_messege:str=""):
    attributes_onnx = op_schema.attributes
    if attribute["name"] not in attributes_onnx:
        warnings.warn(f"\n{start_error_messege} the attribute doenst apply for the operator {op_schema.name} {LINE_WITH_TAB} Attribute Could be Removed")
        return False
    onnx_attribute_type =  attributes_onnx[attribute["name"]].type
    if enum_type.name == onnx_attribute_type.name and enum_type.value==onnx_attribute_type.value:
        return True
    else:
        warnings.warn(f"\n{start_error_messege} the attribute type is invalid for the operator {op_schema.name} {LINE_WITH_TAB} Attribute Could be Removed")
        return False
    
def check_func_op_attributes(enum_type,attribute,func_schema,start_error_messege:str=""):
    attributes_func=[]
    attributes_proto_func = dict()
    if "attribute" in func_schema:
        attributes_func.extend(func_schema["attribute"])
    if "attributeProto" in func_schema:
        attributes_proto_func=  {item["name"]: item for item in func_schema["attributeProto"]}    
        attributes_func.extend(attributes_proto_func.keys())
        
        
    if attribute["name"] not in attributes_func:
        warnings.warn(f"\n{start_error_messege} the attribute ({attribute["name"]}) doenst apply for the function {func_schema["name"]} {LINE_WITH_TAB} Attribute Could be Removed")
        return False

    if attribute["name"] in attributes_proto_func.keys():
        #TODO types need to be the same???
       #attribute_data = attributes_proto_func[attribute["name"]]
       #att_type = check_att_enum(attribute_data["type"])
       #same = att_type == enum_type
       pass
   
    return True
    


def check_attribute_type(type,attribute,op_schema=None,func_schema=None,start_error_messege:str=""):
    enum = check_att_enum(type)
    if enum==None:
        warnings.warn(f"\n{start_error_messege} its type is not a valid AttrType {LINE_WITH_TAB} Attribute Could be Removed")
        return None
    
    if op_schema:
       correct = check_onnx_op_attributes(enum,attribute,op_schema,start_error_messege=start_error_messege)
       if correct:
           return enum
       else:
           return None
    if func_schema:
        
        correct = check_func_op_attributes(enum,attribute,func_schema,start_error_messege=start_error_messege)
        if correct:
            return enum
        else:
            return None
    return enum

def check_attribute_fields(attribute,start_error_messege:str=""):
   
    if not isinstance(attribute,dict):
        warnings.warn(f"\n{start_error_messege} its not a dict {LINE_WITH_TAB} Attribute Could be Removed")
        return False

    
    if 'name' not in attribute:
        warnings.warn(f"\n{start_error_messege} name is missing {LINE_WITH_TAB} Attribute Could be Removed")
        return False
    
    if not isinstance(attribute["name"],str):
        warnings.warn(f"\n{start_error_messege} name has a wrong format {LINE_WITH_TAB} Attribute Could be Removed")
        return False
    
    if "type" not in attribute:
        warnings.warn(f"\n{start_error_messege} type is missing {LINE_WITH_TAB} Attribute Could be Removed")
        return False
    
    if not (isinstance(attribute["type"],str) or isinstance(attribute["type"],int)):
        warnings.warn(f"\n{start_error_messege} type has a wrong format {LINE_WITH_TAB} Attribute Could be Removed")
        return False
    
    
    return True




def repair_attribute_data_value(value,enum_type,start_error_messege,is_list,list_pos,element_id="",subgraph_data=None):
    
    if enum_type in INT_TYPES:
        try: 
            if isinstance(value,str) or isinstance(value,float):
                fix_value=int(value)
                return fix_value
        except Exception:
            _=True
            
        if isinstance(value,int):
            return value
            
        if is_list:
            warnings.warn(f"\n{start_error_messege} The attribute of type {enum_type.name} at the pos {list_pos} of the list must be a integer {LINE_WITH_TAB} Attribute Could be Removed"  )
        else:
            warnings.warn(f"\n{start_error_messege} The attribute of type {enum_type.name} must be a integer {LINE_WITH_TAB} Attribute Could be Removed"  )
        return None
    if enum_type in FLOAT_TYPES:
       
        try: 
            if isinstance(value,str):
                fix_value=float(value)
                return fix_value
        except Exception:
            _=True
        
        if isinstance(value,float):
            return value
        elif isinstance(value,int):
            return float(value)
        
        
        if is_list:
            warnings.warn(f"\n{start_error_messege} The attribute of type {enum_type.name} at the pos {list_pos} of the list must be a float(decimal) {LINE_WITH_TAB} Attribute Could be Removed"  )
        else:
            warnings.warn(f"\n{start_error_messege} The attribute of type {enum_type.name} must be a float (decimal) {LINE_WITH_TAB} Attribute Could be Removed"  )
        return None
    if enum_type in STRING_TYPES:
        if isinstance(value,str):
          return value
        else:
            if is_list:
                warnings.warn(f"\n{start_error_messege} The attribute of type {enum_type.name} at the pos {list_pos} of the list must be a string {LINE_WITH_TAB} Attribute Could be Removed"  )
            else:
                warnings.warn(f"\n{start_error_messege} The attribute of type {enum_type.name} must be a string {LINE_WITH_TAB} Attribute Could be Removed")
            return None
    if enum_type in TENSOR_TYPES:
        if is_list:
            check = check_tensor_complex(value,f"{start_error_messege} The attribute of type {enum_type.name} of the list at pos {list_pos} has a wrong format -> Attribute Could be Removed {LINE_WITH_TAB}",check_storage_data=True,tensor_id=f"{element_id}-idx{list_pos}")
        else:
            check = check_tensor_complex(value,f"{start_error_messege} The attribute of type {enum_type.name} has a wrong format -> Attribute Could be Removed {LINE_WITH_TAB}",check_storage_data=True,tensor_id=element_id)
        if check:
            return value
        else:
            
            return None 
    if enum_type in SPARSE_TYPES:
        if is_list:
            check = check_sparse_complex(value,f"{start_error_messege} The attribute of type {enum_type.name} of the list at pos {list_pos} has a wrong format -> Attribute Could be Removed {LINE_WITH_TAB}",check_storage_data=True,tensor_id=f"{element_id}-idx{list_pos}")
        else:
            check = check_sparse_complex(value,f"{start_error_messege} The attribute of type {enum_type.name} has a wrong format -> Attribute Could be Removed {LINE_WITH_TAB}",check_storage_data=True,tensor_id=element_id)
        if check:
            return value
        else:
            return None
    if enum_type in GRAPH_TYPES:
        if is_list:
            check = check_subgraph(value,subgraph_data,start_error_messege=f"{start_error_messege} The attribute of type {enum_type.name} of the list at pos {list_pos} has a wrong format -> Attribute Could be Removed {LINE_WITH_TAB}",element_id=f"{element_id}-idx{list_pos}")
        else:
            check = check_subgraph(value,subgraph_data,start_error_messege=f"{start_error_messege} The attribute of type {enum_type.name} has a wrong format -> Attribute Could be Removed {LINE_WITH_TAB}",element_id=element_id)
        if check:
            return value
        else:
            return None
    if enum_type in TYPE_PROTO_TYPES:
        if is_list:
            check = check_type_proto(value,f"{start_error_messege} The attribute of type {enum_type.name} of the list at the pos {list_pos} has a wrong format -> Attribute Could be Removed {LINE_WITH_TAB}",element_id=f"{element_id}-idx{list_pos}")
        else:
            check = check_type_proto(value,start_error_messege=f"{start_error_messege} The attribute of type  {enum_type.name} has a wrong format -> Attribute Could be Removed {LINE_WITH_TAB}",element_id=element_id)
        if check:
            return value
        else:
            return None
        
    return None
    

def repair_attribute_data(attribute,enum_type:AttrType,start_error_messege,element_id="",subgraph_data=None):

    correct_all=True
    
    field_name = MAPP_ENUM_LABEL[enum_type]
    if "v" in attribute:
        warnings.warn(f"\n{start_error_messege} fields with name \"v\" are deprecated {LINE_WITH_TAB} Attribute Could be Removed")
        return False
   
    if field_name not in attribute:
        warnings.warn(f"\n{start_error_messege} The attribute of type {enum_type.name} must have the field with name {field_name} {LINE_WITH_TAB} Attribute Could be Removed")
        return False 
       
    is_list=True
    
    if enum_type not in TYPED_LISTS:
        values=[attribute[field_name]]
        is_list=False
        
    else:
        values=attribute[field_name]
        attribute["is_list"]=""

    element_id_type=f"{element_id}-{enum_type.name}"
    
    for idx,value in enumerate(values):
        if is_list:
            element_id_value=f"{element_id_type}-idx{idx}"
        else:
            element_id_value=element_id_type
            
            
        value = repair_attribute_data_value(value,enum_type,start_error_messege,is_list,idx,element_id=element_id_value,subgraph_data=subgraph_data)
        if value==None:
            correct_all=False
        if is_list:
            new_value={"value":value}
            new_value["index"]=idx
            if idx < len(values)-1:
                new_value["next_index"]=idx+1
            if idx == len(values)-1:
                new_value["last_index"]=""
            new_value["type"]=enum_type.name
            
            new_value["element_id"]=element_id_type
            values[idx]=new_value
        else:
            values[idx]=value
    if not correct_all:
        
        return False    

    if not is_list:
        attribute["value"]=values[0]
        del attribute[field_name]
    elif enum_type==AttrType.TYPE_PROTOS:  
        attribute["no-typed_list"]=values
        del attribute[field_name]
    else:
        attribute["typed_list"]=values
        del attribute[field_name]
    return True





    


def repair_attribute(attribute,op_schema=None,func_schema=None,start_error_messege:str="",element_id="",node_id="",subgraph_data=None):
    
    
    enum_type = check_attribute_type(attribute["type"],attribute,op_schema,func_schema,start_error_messege=start_error_messege)
    if enum_type:   
        
        correct = repair_attribute_data(attribute,enum_type,start_error_messege,element_id=element_id,subgraph_data=subgraph_data)
            
        if "docString" in attribute and not isinstance(attribute["docString"],str):
            warnings.warn(f"\n{start_error_messege} The \"docString\" is not a string. Removing \"docString\"")
            attribute["docString"]=""
        if not correct:

            return False,None
    else:
        
        return False,None
    remove_extra_params(attribute,["name","attrType","value","typed_list","no-typed_list","type","docString","element_id","is_list"],start_error_messege)    
    attribute["element_id"]=element_id
    attribute["node_id"]=node_id
    return True,attribute


def check_missing_onnx_att(att_names,op_schema):
    for att in op_schema.attributes.keys():
        if att not in att_names and op_schema.attributes[att].required :
            return att
    return None

def check_missing_atts_list(att_names,correct_list):
    for att in correct_list:
        if att not in att_names:
            return att
    return None

        
def check_missing_attributes(att_names,op_schema=None,func_schema=None,start_error_messege:str=""):
    
    
    if op_schema:
        att =check_missing_onnx_att(att_names,op_schema) 
        if att:
            warnings.warn(f"\n{start_error_messege} its missing the attribute {att} for the operator {op_schema.name}")
            return False
    if func_schema and "attribute" in func_schema:
        
        att = check_missing_atts_list(att_names,func_schema["attribute"])
        if att:
            warnings.warn(f"\n{start_error_messege} its missing the attribute {att} for the function {func_schema["name"]}")
            return False
    return True
def get_op_schema_func(node,func_data,type_operator_data):
    op_schema=None
    func_schema=None
    if type_operator_data and node and type_operator_data[1]:
        # obtenemos informacion sobre si los atributos son opcionales
        op_schemas = find_operators(node["opType"],node["domain"],node["version"])
        if len(op_schemas)>0:
            op_schema=op_schemas[0]
        else:
            raise RuntimeError(f"Values of domain ({node["domain"]}),name ({node["opType"]}),version ({node["version"]}) are not operators of the model")
         
   
    if type_operator_data and func_data and type_operator_data[0]:
        # get function and get attribute and attribute_proto
        
        funcs = function_data_find(func_data,node["domain"],node["opType"],node["overload"])
        if len(funcs)>0:
            func_schema=funcs[0]
        else:
            raise RuntimeError(f"Values of domain ({node["domain"]}),name ({node["opType"]}),overload ({node["overload"]}) are not functions of the model")
  
    
    return op_schema,func_schema
    
def check_attributes(attributes,node=None,type_operator_data=None,func_data=None,element_id="",start_error_messege:str="",subgraph_data=None):
    
    correct_all=True
    correct=True
    op_schema,func_schema = get_op_schema_func(node,func_data,type_operator_data)  
        
    attribute_names=[]
    incorrect_value_names=[]
    
    if not isinstance(attributes,list):
        warnings.warn(f"\nThe attributes of the node ({node["name"]}) has a wrong format: It is not a list {LINE_WITH_TAB} Attribute Could be Removed")
        return False,False
    
    for idx,attribute in enumerate(attributes):
        att_name=""
        if "name" in attribute:
            att_name = attribute["name"]
        if att_name=="":
            start_error_messege=f"{start_error_messege} The attribute entry at pos {idx} has a wrong format {LINE_WITH_TAB}"
        else:
            start_error_messege=f"{start_error_messege} The attribute entry at pos {idx} and name ({att_name}) has a wrong format {LINE_WITH_TAB}"  
        correct = check_attribute_fields(attribute,start_error_messege)
        
        
        if correct:
            
            
                
            att_element_id = f"{element_id}-Attribute-{att_name}"
            node_id = element_id
            
            correct_att,new_attr = repair_attribute(attribute,op_schema,func_schema,start_error_messege=start_error_messege,element_id=att_element_id,node_id=node_id,subgraph_data=subgraph_data) 
            
            
        if correct and correct_att:
            attributes[idx]=new_attr
            attribute_names.append(attribute["name"])
        else:
            
            if not correct_att and correct:
                
                incorrect_value_names.append(attribute["name"])
            correct_all=False
            del attributes[idx]
            
    
    # Add default values if possible
    if func_schema!=None and "attributeProto" in func_schema:
        for idx,attribute in enumerate(func_schema["attributeProto"]):
            if attribute["name"] not in attribute_names:
                attributes.append(attribute)
        
    
            
    if node:
        
        all_names = attribute_names.copy()
        all_names.extend(incorrect_value_names)
        
        correct = check_missing_attributes(all_names,op_schema,func_schema,f"The node ({node["name"]}) ")
        
    
    return correct,correct_all

      
    #las funciones definen atributos con attributeproto


import os
import json


def get_log_paths(log_dirs,log_file_name,work_dir):
    log_paths=[]
    if not isinstance(log_dirs,list):
        log_dirs=[log_dirs]
        
    for log_dir in log_dirs:
        
        if os.path.isabs(log_dir):
            log_path = os.path.join(log_dir,log_file_name)
        else:
            log_path = os.path.join(work_dir,log_dir,log_file_name)
        log_paths.append(log_path)
    return log_paths

def save_json(json_file,path):

    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path,"w") as report:
        json.dump(json_file,report,indent=4,default=str)
    
def load_json(path):
    if os.path.exists(path):
        with open(path,"r") as report:
         return json.load(report)
    raise ValueError(f"Path {path} doesnt exit")
    
def remove_files(folder):
    for filename in os.listdir(folder):
        file_path = os.path.join(folder, filename)
        if os.path.isfile(file_path):
            try:
                os.remove(file_path)
            except Exception:
                pass



import os
from json_util import save_json,load_json,get_log_paths
from util_process.util import remove_extra_params

from util_process.operators import find_operators,get_key_operator,is_domain_ONNX,get_correct_status
from util_process.operators import is_operator_func,get_operator_url,is_operator_shape_inference,is_onnx_runtime
from onnx.defs import  OpSchema
from warnings_thread_safe import warnings 
from repair_functions.repair_opsets import repair_opset_and_funcs

from util_process.data import get_opset_data,function_data_find,opset_data_find,get_functions_data,function_key,opset_key,get_opset_values,get_func_values
from repair_nodes.repair_nodes_base import repair_nodes

from repair_types.repair_initializer import repair_initializers
from repair_types.reapair_value_info import revise_value_infos_graph
from repair_types.repair_tensor import DataType,get_label_data_type
from util_process.util import LocationType,OperatorStatus,get_location_label,get_operator_status_label

from repair_types.repair_metadata_prop import check_metadata_entries
from util_process.util import LINE_WITH_TAB,create_unique_names

from repair_types.repair_annotation import check_annotations







def __check_graph_fields__(graph):
    if "node" not in graph or not isinstance(graph["node"],list):
            graph["node"]=[]
            warnings.warn("Graph has nodes missing or are not a list, the result model will have empty nodes. ")
    if "initializer" not in graph or not isinstance(graph["initializer"],list):
            graph["initializer"]=[]      
    if "sparseInitializer" not in graph or not isinstance(graph["sparseInitializer"],list):
            graph["sparseInitializer"]=[]

    if "name" in graph :
        base_name =graph["name"]
    else:
        base_name="unnamed_graph"   
        
    graph["name"] =create_unique_names([],base_name)
    
        
            
    return True
       
                
def __funct_contains_name__(data,name):
    for function in data["functions"]:
        
        if "name" in function and function["name"]==name:
            return True
    return False

def __is_operator_depracted_on_version__(operator,version):
    new_ops:list[OpSchema] = find_operators(operator.name,operator.domain,version)
    if len(new_ops)==0:
        return False
    return new_ops[0].deprecated
 
def __get_version_opset_to_operator__(import_data,domain,operator:OpSchema):
    opsets = opset_data_find(import_data,domain)
    
    best_version=None
    for opset in opsets:
        version=opset["version"]
        if int(version)>=int(operator.since_version) and (best_version==None or int(version)>int(best_version)): 
            if not __is_operator_depracted_on_version__(operator,version):
                best_version=version    
            
    return best_version

                         
def __add_operator_if_onnx__(name,domain,version,dict_ops:dict,import_data):
    key = get_key_operator(domain,name,version)
    
    if key not in dict_ops.keys():
        operator = find_operators(name=name,domain=domain,version=version)
        
        if len(operator)>0:
            dict_ops[key]={"schema":operator[0],"opset_version":__get_version_opset_to_operator__(import_data,domain,operator[0])}
            
            
def add_entry_if_exists(dict_instance,key,entity,new_key=None):
    value=None
    if isinstance(entity,dict) and key in entity.keys():
        value = entity[key]
    if isinstance(entity,object) and hasattr(entity, key):
        value = getattr(entity,key)
    if value!=None and not new_key:
        dict_instance[key]=str(value)
    if value!=None and new_key:
        dict_instance[new_key]=str(value)

def __fill_status_codes(data):
    
    data["operator_types"]=[]
    type_list=data["operator_types"]
    for tensor_type in OperatorStatus:
        label = get_operator_status_label(tensor_type)

        type_list.append({"prefLabel":tensor_type.name,"altLabel":tensor_type.value,"label":label})
    
    data["location_types"]=[]
    type_list=data["location_types"]
    for tensor_type in LocationType:
        label = get_location_label(tensor_type)

        type_list.append({"prefLabel":tensor_type.name,"altLabel":tensor_type.value,"label":label})
    
    
    data["tensor_types"]=[]
    type_list=data["tensor_types"]
    for tensor_type in DataType:
        label = get_label_data_type(tensor_type)

        type_list.append({"prefLabel":tensor_type.name,"altLabel":tensor_type.value,"label":label})
        
    data["map_key_types"]=[]
    type_list=data["map_key_types"]
    
    map_key_types=[DataType.UINT64,DataType.UINT32,DataType.UINT16,DataType.UINT8,
                   DataType.INT64,DataType.INT32,DataType.INT16,DataType.INT8,DataType.STRING]
    
    for tensor_type in map_key_types:
        label = get_label_data_type(tensor_type)

        type_list.append({"prefLabel":tensor_type.name,"altLabel":tensor_type.value,"label":label})
 

 
            
def __fill_operator_metadata__(data):
    
    operators_onnx={}
    import_data=get_opset_data(data["opsetImport"])
    for node in data["graph"]["node"]:
        
        if "opType" in node and "domain" in node and "version" in node :
            __add_operator_if_onnx__(node["opType"],node["domain"],node["version"],operators_onnx,import_data)
        
    for func in data["functions"]:
        import_data=get_opset_data(func["opsetImport"])
        for node in func["node"]:
            if "opType" in node and "domain" in node and "version" in node :
                __add_operator_if_onnx__(node["opType"],node["domain"],node["version"],operators_onnx,import_data)
    data["operators"]=[]
                
    for operator_data in operators_onnx.values():
        operator=operator_data["schema"]
        operator_entry=dict()
        add_entry_if_exists(operator_entry,"name",operator)
        add_entry_if_exists(operator_entry,"domain",operator)
        add_entry_if_exists(operator_entry,"since_version",operator)
        add_entry_if_exists(operator_entry,"doc",operator,new_key="docString")
        
        #operator status
        if hasattr(operator,"support_level"):
            operator_entry["status"]=get_correct_status(operator.support_level.name)

        operator_entry["is_function"]=is_operator_shape_inference(operator)
        operator_entry["is_shape_inference"]=is_operator_func(operator)
        if operator_entry["domain"]!=None and operator_entry["name"] and (is_domain_ONNX(operator_entry["domain"]) or is_onnx_runtime(operator_entry["domain"])):
            
            
            url = get_operator_url(operator_entry["domain"],operator_entry["name"])
            if url:
                operator_entry["url"]=url
        
        
        
        opset_id =f"v{operator_data["opset_version"]}"
        if operator_entry["domain"]!="":
            opset_id=f"domain-{operator_entry["domain"]}-{opset_id}"
        operator_entry["opset_id"]=opset_id
        
        #TODO: get default_attributes
        #TODO: get input/output constrains
        
        
        element_id=f"name-{operator_entry["name"]}"
        if operator_entry["domain"]!="":
            element_id=f"{element_id}-domain-{operator_entry["domain"]}"
        element_id=f"{element_id}-v{operator_entry["since_version"]}"
        operator_entry["element_id"]=element_id
        
        data["operators"].append(operator_entry)
    
                
 
 

 
def check_first_fields(data):
    
    if "graph" not in data or not isinstance(data["graph"],dict):
        warnings.warn("Graph was missing from file or the format was wrong, the new graph will be empty. ")
        data["graph"]=dict()
    
    if "functions" not in data or not isinstance(data["functions"],list):
        data["functions"]=[]
        
    if "opsetImport" not in data or not isinstance(data["opsetImport"],list):
        data["opsetImport"]=[]
    return True
 
 



            

 


        
# Main library function of the pre_process function
# Inputs json dict and outputs repair/modified json_dict for rdf mapping
def pre_process(data:dict):
    
    check_first_fields(data)
    repair_opset_and_funcs(data)
    
    __check_graph_fields__(data["graph"])
    
    repair_initializers(data["graph"])
    revise_value_infos_graph(data["graph"])
    
    repair_nodes(data)
    data["graph"]["element_id"]=""
    
    check_annotations(data["graph"],element_id="",start_error_messege= f"\n The quantizationAnnotation of graph ({data["graph"]["name"]}) has a wrong format {LINE_WITH_TAB}")
            
    remove_extra_params(data["graph"],["input","output","node","name","initializer","sparseInitializer","valueInfo","element_id","metadataProps","docString","quantizationAnnotation"],
                        start_error_messege=f"The Main graph with name ({data["graph"]["name"]})")
    remove_extra_params(data["graph"],["input","output","node","name","initializer","sparseInitializer","valueInfo","element_id","metadataProps","docString","quantizationAnnotation"],
                        start_error_messege="The model")
    check_metadata(data)
    
    
    __fill_operator_metadata__(data)
    __fill_status_codes(data)
    
def check_metadata(data):
    
    if "metadataProps" in data and data["metadataProps"]!=[]:
        
        correct_meta,_ =check_metadata_entries(data["metadataProps"],start_error_messege=f"\nThe file has a wrong metadata_prop param {LINE_WITH_TAB}",keys=[],element_id="")

        if not correct_meta:
            data["metadataProps"]=[] 
          
            
    if "metadataProps" in data["graph"] and data["graph"]["metadataProps"]!=[]:
        correct_meta,_ =check_metadata_entries(data["graph"]["metadataProps"],keys=[],element_id=f"Graph-{data["graph"]["name"]}",
            start_error_messege=f"\nThe graph {data["graph"]["name"]} has a wrong metadata_prop param {LINE_WITH_TAB}"
                                               )

        if not correct_meta:
            data["graph"]["metadataProps"]=[]  
    
    
from datetime import datetime  

# Main cmd function of the pre_process function
# Inputs json dict and uses parms for managing cache files,log dirs, etc
def __pre_process__(data,relative_dir="",tmp_dir="tmp",log_dirs="logs",cache=True):
    
    cache_executed=False
    file_name="preprocess_model.json"
        
    tmp_path = os.path.join(relative_dir,tmp_dir,file_name)
    if not isinstance(log_dirs,list):
        log_dirs=[log_dirs]
    
    log_paths=log_paths=get_log_paths(log_dirs,file_name,relative_dir)
    
        
    is_in_cache = os.path.exists(tmp_path)
    
    
    if cache and is_in_cache:
        cache_executed=True
            
    if not cache or (not is_in_cache):
        
        pre_process(data) 
        for log_path in log_paths:
            save_json(data,log_path)
        
        save_json(data,tmp_path)     
                   
    return tmp_path,cache_executed



from onnx import defs

import onnxruntime.capi.onnxruntime_pybind11_state as rtpy
from onnx.defs import  OpSchema
from . import ONNX_DOMAIN



OPERATORS: dict[str, OpSchema] = dict()
ONNX_RUNTIME_OP: dict[str, OpSchema] = dict()
DOMAINS = []
ONNX_PREDEFINED_DOMAINS=[defs.ONNX_DOMAIN,defs.ONNX_ML_DOMAIN,defs.AI_ONNX_PREVIEW_TRAINING_DOMAIN]
MICROSOFT_COM_DOMAINS = ["com.microsoft","com.microsoft.experimental","com.microsoft.nchwc"]
ONNX_PREDEFINED_DOMAINS.extend(MICROSOFT_COM_DOMAINS)



DOMAINS_VERSION_ONE=[defs.AI_ONNX_PREVIEW_TRAINING_DOMAIN]
DOMAINS_VERSION_ONE.extend(MICROSOFT_COM_DOMAINS)


DEPRACTED: dict[str, OpSchema] = dict()
from .data import __dynamic_dict_find__



from util_process.util import OperatorStatus,get_operator_status_label



def get_operator_url(domain,name):
  is_runtime = is_onnx_runtime(domain,name)
  
  if is_domain_ONNX(domain) and not is_runtime:
    return f"https://onnx.ai/onnx/operators/onnx_{domain}_{name}.html"
  
  elif is_runtime:
    if domain in MICROSOFT_COM_DOMAINS:
      return f"https://github.com/microsoft/onnxruntime/blob/main/docs/ContribOperators.md#{domain}.{name}"
    return None
  else:
    return None


# is the domain,name,version introduced by runtime
def is_onnx_runtime(domain,name=None,version=None):
  list_op = find_operators(name,domain,version)
  if len(list_op)==0:
    return False
  
  if domain and name==None and is_domain_ONNX(domain):
    return False
  list_run = find_operators(name,domain,version,ONNX_RUNTIME_OP)
  if len(list_run) ==0:
    return False
  # los operadores definidos por runtime solo se añaden si no se han añadido en onnx normal
  return True


def get_key_operator(domain,name,version):
  return f"\"{domain}\".\"{name}\".\"{version}\""


def get_correct_status(label):
  if label=="COMMON" or label=="STABLE":
    new_label="STABLE"
    label_code=1
  else:
    new_label="EXPERIMENTAL"
    label_code=0
  label = get_operator_status_label(OperatorStatus[new_label])
  return{"prefLabel":new_label,"altLabel":label_code,"rdfLabel":label}


for schema in defs.get_all_schemas_with_history():
  key = get_key_operator(schema.domain,schema.name,schema.since_version)
  if key not in OPERATORS.keys():
    if schema.deprecated:
      DEPRACTED[get_key_operator(schema.domain,schema.name,schema.since_version)]=schema
    OPERATORS[get_key_operator(schema.domain,schema.name,schema.since_version)]=schema
  if schema.domain not in DOMAINS:
      DOMAINS.append(schema.domain)
      
for schema in rtpy.get_all_operator_schema():
  key = get_key_operator(schema.domain,schema.name,schema.since_version)
  if key not in OPERATORS.keys():
    if schema.deprecated:
      DEPRACTED[get_key_operator(schema.domain,schema.name,schema.since_version)]=schema
    OPERATORS[get_key_operator(schema.domain,schema.name,schema.since_version)]=schema
    ONNX_RUNTIME_OP[get_key_operator(schema.domain,schema.name,schema.since_version)]=schema
  if schema.domain not in DOMAINS:
      DOMAINS.append(schema.domain)



def is_deprecated(domain,name,version):
    return get_key_operator(domain,name,version) in DEPRACTED.keys()


def is_domain_ONNX(domain:str):
    """
    The function `is_domain_ONNX` checks if a given domain is in the predefined ONNX domains.
    
    Args:
      domain (str): The `is_domain_ONNX` function is checking if a given domain is in the
    `ONNX_PREDEFINED_DOMAINS` list. The function takes a `domain` parameter of type string and returns a
    boolean value indicating whether the domain is in the predefined list of domains for ONNX.
    
    Returns:
      The function `is_domain_ONNX` is checking if the input `domain` is in the list of predefined
    domains `ONNX_PREDEFINED_DOMAINS` and returning a boolean value based on that check.
    """
    return domain in ONNX_PREDEFINED_DOMAINS

def is_register_domain(domain:str):
    """
    The function `is_register_domain` checks if a given domain is in a predefined list of domains.
    
    Args:
      domain (str): The `is_register_domain` function takes a domain name as input and checks if it is
    in the `DOMAINS` list. If the domain is found in the list, the function will return `True`,
    indicating that the domain is a registered domain. If the domain is not in the list,
    
    Returns:
      The function `is_register_domain` is returning a boolean value indicating whether the input domain
    is in the `DOMAINS` list.
    """
    
    return domain in DOMAINS

def get_domain_ONNX_version(domain:str):
    """
    This function returns the ONNX opset version based on the input domain.
    
    Args:
      domain (str): The function `get_domain_ONNX_version` takes a `domain` parameter as input. The
    function checks the value of the `domain` parameter and returns the corresponding ONNX version based
    on the domain provided. The possible values for the `domain` parameter are `defs.ONNX_DOMAIN`, `defs
    
    Returns:
      The function `get_domain_ONNX_version` returns the ONNX opset version based on the input domain
    provided. If the domain is `defs.ONNX_DOMAIN`, it returns the ONNX opset version. If the domain is
    `defs.ONNX_ML_DOMAIN`, it returns the ONNX-ML opset version. If the domain is
    `defs.AI_ONNX_PREVIEW_TRAINING_DOMAIN`, it
    """
    if domain == defs.ONNX_DOMAIN:
       return  defs.onnx_opset_version()
    if domain == defs.ONNX_ML_DOMAIN:
       return  defs.onnx_ml_opset_version()
    if domain in DOMAINS_VERSION_ONE:
       return  1
    
   

def is_opertor_list_func_type(operators:dict[str, OpSchema]=OPERATORS):
    """
    This function checks if a list of operators contains at least one valid operator function.
    
    Args:
      operators (list[OpSchema]): The `operators` parameter in the `is_opertor_list_func_type` function
    is expected to be a list of `OpSchema` objects. The function checks if the list is empty, if it
    contains only one element, or if any of the elements in the list are of type `Op
    
    Returns:
      The function `is_opertor_list_func_type` returns a tuple containing two values:
    1. A boolean value indicating whether there are any valid operators in the input list `operators`.
    2. A list of valid operators that were filtered from the input list.
    """
    if len(operators)==0:
        return False,[]
    if len(operators)==1:
        return is_operator_func(operators[0])
    filter_operators = []
    for operator in operators:
        if is_operator_func(operator):
            filter_operators.append(operator)
            
    return len(filter_operators)>0,filter_operators  
    
# busqueda dinamica en un diccionario con claves compuestas
def find_operators(name:str=None,domain:str=None,version:str=None,operator_list:dict[str, OpSchema]=OPERATORS):
    """
    The function `find_operators` searches for operators based on specified criteria such as name,
    domain, and version in a given list of operator schemas.
    
    Args:
      name (str): The `name` parameter is a string that represents the name of an operator. It is used
    to filter the list of operators based on the specified name. If provided, only operators with a
    matching name will be included in the result.
      domain (str): The `domain` parameter in the `find_operators` function is used to filter operators
    based on their domain. If you provide a value for the `domain` parameter, the function will only
    return operators that belong to that specific domain. If you don't provide a value for `domain`, the
      version (str): The `version` parameter in the `find_operators` function is used to filter the
    operators based on a specific version. If you provide a version value when calling the function, it
    will only return operators that match that version. If you don't provide a version value, it will
    not filter based
      operator_list (list[OpSchema]): The `operator_list` parameter in the `find_operators` function is
    expected to be a dictionary where the keys are strings in the format "domain.name.version" and the
    values are instances of `OpSchema`. The function iterates over this dictionary and filters out the
    operators based on the provided `
    
    Returns:
      The function `find_operators` returns a list of `OpSchema` objects that match the criteria
    specified by the `name`, `domain`, and `version` parameters.
    """
    if name!=None and domain!=None and version!=None:
      key = get_key_operator(domain,name,version)
      if key in operator_list.keys():
        return [operator_list[key]]
      else:
        return []
    return __dynamic_dict_find__(operator_list,3,[domain,name,version])
    


def is_operator_func(operator:OpSchema):
    """
    The function `is_operator_func` checks if an operator has a context-dependent function or a regular
    function.
    
    Args:
      operator (OpSchema): OpSchema - an object representing an operator in a computational graph, with
    attributes like has_context_dependent_function and has_function.
    
    Returns:
      The function `is_operator_func` is returning a boolean value based on the conditions provided. It
    checks if the input `operator` has either a context-dependent function or a regular function, and
    returns `True` if either condition is met, otherwise it returns `False`.
    """
    if not hasattr(operator,"has_context_dependent_function") or not hasattr(operator,"has_function"):
      return False

    return operator.has_context_dependent_function or operator.has_function
  
def is_operator_shape_inference(operator:OpSchema):
  '''The function `is_operator_shape_inference` checks if an operator has a type and shape inference
  function.
  
  Parameters
  ----------
  operator : OpSchema
    The function `is_operator_shape_inference` takes an input parameter `operator` of type `OpSchema`.
  The function checks if the `operator` has a type and shape inference function defined and returns a
  boolean value based on that check.
  
  Returns
  -------
    The function `is_operator_shape_inference` is checking if the input `operator` has a type and shape
  inference function. It returns a boolean value indicating whether the `operator` has a type and
  shape inference function (`True`) or not (`False`).
  
  '''
  return operator.has_type_and_shape_inference_function



# existe algun operator registrado con ese nombre/dominio/version  que su dominio sea ONNX
def is_onnx_operator(name:str=None,domain:str=None,version:str=None,operator_list:dict[str, OpSchema]=OPERATORS):
    """
    This function checks if an ONNX operator exists based on its name, domain, and version.
    
    Args:
      name (str): The `name` parameter in the `is_onnx_operator` function is used to specify the name of
    the ONNX operator you are checking for.
      domain (str): The `domain` parameter in the `is_onnx_operator` function is used to specify the
    domain of the ONNX operator. The ONNX specification organizes operators into different domains based
    on their functionality or area of application. By providing a domain value, you can filter the
    operators based on that specific
      version (str): The `version` parameter in the `is_onnx_operator` function is used to specify the
    version of the ONNX operator. This parameter allows you to filter operators based on their version
    numbers. If a specific version is provided, only operators with a version greater than or equal to
    the specified version will
      operator_list (list[OpSchema]): The `operator_list` parameter in the `is_onnx_operator` function
    is a list of `OpSchema` objects. These objects likely contain information about ONNX operators such
    as their name, domain, version, and other relevant details. The function uses this list to find and
    filter operators based on
    
    Returns:
      The function `is_onnx_operator` returns a tuple containing two elements:
    1. A boolean value indicating whether there are any ONNX operators that match the specified criteria
    (name, domain, version).
    2. A list of ONNX operators that match the specified criteria (if any), filtered based on the
    provided name, domain, and version.
    """
    
    
    
    if domain and (domain not in ONNX_DOMAIN):
        return False,[]
    
    operators = find_operators(name,domain,version,operator_list)
    if len(operators)==0:
        return False,[]
    filter_operators = []
    for operator in operators:
        if operator.domain in ONNX_DOMAIN and (domain==None or domain==operator.domain) and (version==None or version>=operator.since_version):
            filter_operators.append(operator)

    
    return len(filter_operators)>0,filter_operators



#devuelve si existe algun operator registrado con ese nombre/dominio/version y la lista de coincidencias
def is_register_operator(name:str=None,domain:str=None,version:str=None,operator_list:dict[str, OpSchema]=OPERATORS):
    """
    This function checks if a given operator is registered based on its name, domain, and version.
    
    Args:
      name (str): The `name` parameter in the `is_register_operator` function is a string that
    represents the name of the operator you want to check for registration.
      domain (str): The `domain` parameter in the `is_register_operator` function refers to the domain
    of the operator being checked. It is used to filter operators based on their domain.
      version (str): The `version` parameter in the `is_register_operator` function is used to specify
    the version of the operator. It is used to filter the operators based on their version number. If a
    version is provided, only operators with a version greater than or equal to the specified version
    will be considered in the
      operator_list (list[OpSchema]): The `operator_list` parameter in the `is_register_operator`
    function is a list of `OpSchema` objects. These objects likely represent different operators used in
    a system or application. The function is designed to check if a specific operator is registered
    based on the provided `name`, `domain`, and
    
    Returns:
      The function `is_register_operator` is returning a tuple containing two values. The first value is
    a boolean indicating whether there are any filtered operators that match the specified criteria, and
    the second value is a list of those filtered operators.
    """
    
    if domain and not is_register_domain(domain):
      return False,[]
    
    operators = find_operators(name,domain,version,operator_list)
    if len(operators)==0:
      return False,[]
    
    filter_operators = []
    for operator in operators:
        if (domain==None or domain==operator.domain) and (version==None or version>=operator.since_version):
            filter_operators.append(operator)
    
    
    return len(filter_operators)>0,filter_operators











import argparse
import inspect
import os
import sys
import re



import traceback
from typing import Union
from enum import Enum
sys.path.append(os.path.abspath(os.path.dirname(__file__)))


from warnings_thread_safe import warnings 

from clean_model import process_model as execute_clean_model
from pre_process_module.pre_process import __pre_process__ as execute_pre_process
  
from ruamel.yaml import YAML
from datetime import datetime
import time
import logging

import functools
from pathvalidate import ValidationError,validate_filepath

from json_util import save_json,get_log_paths,remove_files


import threading

LOADED_CACHE = "loaded_model.json"


VALID_CACHE_PARTS=["load-model","pre-process","yamml2rml","mapping","all","None"]
DEFAULT_RESOURCE_BASE_URL="http://base.onnx.model.com/resource/"

DEFAULT_ONNX2RDF_URI="http://base.onnx.model.com#"
DEFAULT_ONNX2RDF_RESOURCE="http://base.onnx.model.com/resource/"


import urllib.request
from pathlib import Path

MAPPER_NAME = "rmlmapper.jar"

JAR_URL = "https://github.com/RMLio/rmlmapper-java/releases/download/v7.3.3/rmlmapper-7.3.3-r374-all.jar"
JAR_PATH = Path(__file__).parent / MAPPER_NAME

def download_rmlmapper_jar():
    if not JAR_PATH.exists():
        print(f"Downloading RMLMapper JAR from {JAR_URL}...")
        JAR_PATH.parent.mkdir(parents=True, exist_ok=True)
        urllib.request.urlretrieve(JAR_URL, JAR_PATH)
        print(f"Downloaded to {JAR_PATH}")
    else:
        print(f"RMLMapper JAR already exists at {JAR_PATH}")
    return JAR_PATH



class ErrorsONNX2RDF(Enum):
    NONE_ERROR=0
    LOADING_MODEL = 1
    PREPROCESING_MODEL=2
    YAMML2RML_COPY=3
    YAMML2RML=4
    MAPPING = 5
    WARNING_ON_PREPROCESS = 6
    FILE_NOT_PRESENT=7
    CACHE_FILE_NOT_PRESENT = 8



def setup_logging(log_files=[], to_console=True,error_file=None,id_process="",log_id=""):
    
    handlers = []
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s") 
    console_handler = logging.StreamHandler(sys.stdout)
   
    if not isinstance(log_files,list):
        log_files=[log_files]
    for log_path in log_files:
        
        os.makedirs(os.path.dirname(log_path),exist_ok=True)
        if log_path is not None:
            handlers.append(logging.FileHandler(log_path, mode='w'))   
    if error_file is not None:
            handlers.append(logging.FileHandler(error_file, mode='w'))    
    if to_console:
            handlers.append(console_handler)
    log_code = "onnx_logger"
    if id_process!="":
        log_code=f"onnx_logger_{id_process}"
    if log_id!="":
        log_code=log_id
    logger = logging.getLogger(log_code)
    
    
    level = logging.DEBUG
    set_handlers_format(handlers,formatter)
    
    logger.setLevel(level)
    
    for handler in handlers:
        logger.addHandler(handler)
    
    return logger
      
def set_handlers_format(handlers,formater):
    for handler in handlers:
        handler.setFormatter(formater) 
def cleanup_handlers(logger):
    if logger==None:
        return None
    
    for handler in logger.handlers[:]:
        handler.close()
        logger.removeHandler(handler)        
 


def combine_work_folder(work_folder,path,file_name=""):
    if not os.path.isabs(path):
        path = os.path.join(work_folder,path)
        
    if file_name!="":
        path = os.path.join(path,file_name)
    return path

def setup_error_file(error_folder,work_dir,model_name_path):

    error_folder = combine_work_folder(work_dir,error_folder)
    folder_model = model_name_path
    specific_error = os.path.join(error_folder,folder_model)
    os.makedirs(specific_error,exist_ok=True)
 
    log_file_name = "logs.log"  
    error_log_file = os.path.join(specific_error,log_file_name)
    
    return specific_error,error_log_file

import os
def remove_empty_dirs(base_path, target_dir):
    base_path = os.path.abspath(base_path)
    target_path = os.path.abspath(os.path.join(base_path, target_dir)) if not os.path.isabs(target_dir) else os.path.abspath(target_dir)
  
    if os.path.commonpath([base_path, target_path]) != base_path:
        raise ValueError(f"Target path '{target_path}' is not within base path '{base_path}'")

    current = target_path
    while current != base_path and os.path.isdir(current) and not os.listdir(current):
        try:
            os.rmdir(current)
        except OSError:
            break  
        current = os.path.dirname(current)

def remove_folder_rec(folder,work_dir,model_name_path):
    folder = combine_work_folder(work_dir,folder)
    specific_model_folder = os.path.join(folder,model_name_path)
    remove_empty_dirs(folder, specific_model_folder)


def copy_edit_mappings(yarrml_mapping,work_dir,new_path,model_full_uri,id_process="",log_dirs="logs",tmp_dir="tmp/mappings",cache=False):
    
    file_name = "modified_mapping"
    
    if id_process!="":    
        mapping_name = file_name+"_"+str(id_process)
    else:
        mapping_name = file_name
    if not isinstance(log_dirs,list):
        log_dirs=[log_dirs]
    log_file_name=  f"{file_name}.yaml"  
    log_paths=get_log_paths(log_dirs,log_file_name,work_dir)

        
    output_yaml_path = os.path.join(work_dir,tmp_dir,mapping_name+".yaml")
    cache_activated=False
    exists_path=os.path.exists(output_yaml_path)
    if cache and exists_path:
        cache_activated=True
    
    
    if not cache or not exists_path:
    
        yaml_dumper=YAML() 
        yaml_dumper.allow_duplicate_keys = True
        
        with open(yarrml_mapping,'r') as f:
            
            yaml_dumper.preserve_quotes = True
            yaml_f = yaml_dumper.load(f)
            
        for source in yaml_f["sources"].values():
            source["access"]=new_path
        
        yaml_f["prefixes"]["modelR"]=model_full_uri
        
        yaml_f["prefixes"]["base"]=DEFAULT_ONNX2RDF_URI
        
        yaml_f["prefixes"]["baseR"]=DEFAULT_ONNX2RDF_RESOURCE
        
        
            
        os.makedirs(os.path.dirname(output_yaml_path), exist_ok=True)

        with open(output_yaml_path, 'w',) as f :
            yaml_dumper.dump(yaml_f,stream=f)
        
        for log_path in log_paths:
            save_json(yaml_f,log_path)
           
    return output_yaml_path,cache_activated


class RDF_formats(Enum):
    NQUADS = 'nquads'
    TURTLE = 'turtle'
    TRIG = 'trig'
    TRIX = 'trix'
    JSONLD = 'jsonld'
    HDT = 'hdt'



FORMATS_EXTENSIONS = {
        'nquads': '.nq',
        'turtle': '.ttl',  
        'trig': '.trig',
        'trix': '.trix',
        'jsonld': '.jsonld',
        'hdt': '.hdt'
    }
def get_extension(output_format:Union[RDF_formats, str]):
   
    if isinstance(output_format,RDF_formats):
        str_format=output_format.value
    else:
        str_format=output_format
        
    return FORMATS_EXTENSIONS.get(str_format, None)


import subprocess


def parse_args():

    parser = argparse.ArgumentParser(description="Process a ONNX file or folder with ONNX files to RDF files.")
    parser.add_argument("model_path", help="Path to the onnx file to process or folder with multiple files. Path can be absolute or relative.")
    parser.add_argument("--target_path", default="rdfs", help="Directory where rdf files will be stored (default: 'rdfs'). Path can be absolute and in case relative path actual folder or --work_folder will be used")
    parser.add_argument("--rdf_format", default="nquads", help="Available rdf formats (nquads (default), turtle, trig, trix, jsonld, hdt).")
    parser.add_argument("--log_persistant", action=argparse.BooleanOptionalAction, help="(default: Off) If set instead of the logs being dump on the last_execution folder, a new folder with timestamps is created. \nIf --log_extra is executed only the folder with time_stamps gets the extra files")
    parser.add_argument("--log_extra", action=argparse.BooleanOptionalAction, help=" (default: Off) If set, temporary files created during the pipeline will be store within the logs in the folder --log_folder\n If If --log_persistant is executed only the folder with time_stamps gets the extra files.")
    parser.add_argument("--log_folder", default="logs", help="Folder for logs (default: 'logs') Path can be absolute or relative.")
    parser.add_argument('--cache',nargs='*',metavar='PART',help=("Enable caching. Temporary files created --tmp_folder will not be removed and used as cache . "
            "If followed by parts, only caches files created at those processes (default part: 'all') (default : 'None'). Valid parts: "
            + ", ".join(VALID_CACHE_PARTS))
    )
    parser.add_argument("--error_folder", default="errors", help="Temporary folder where files/info used to recreate error.")
    
    
    parser.add_argument("--debug", action=argparse.BooleanOptionalAction, help="If set, instead of temporary files being delted if not otherwise request by cache those files would be stored .")

    parser.add_argument("--tmp_folder", default="tmp", help="Temporary folder where files are created for pipeline execution. Path can be absolute or relative.")
    parser.add_argument("--id_process", default="", help="If set, this id will be added to files created for parallel processing, just for rml mapping.")
    parser.add_argument("--work_folder", default="", help="Change the relative folder for searching models, creating logs folder or rdf folders.")
    parser.add_argument("--stop-parsing",action=argparse.BooleanOptionalAction, help="If set only preprocessing.")
    parser.add_argument("--model_name_path",default=None, help="Model Name use for the URI of the model (\"{base_url}{model_name_path}\"), (default: 'model_path: filename').")
    parser.add_argument("--base_url", default=None, help="Base url of the resources of the model folowing. The URI of the resources will be (\"{base_url}{model_name_path}\"), (default: 'http://base.onnx.model.com/resource/{model_name_path}').")
    parser.add_argument('-v', '--verbose', action=argparse.BooleanOptionalAction,help="The rml mapping will output in verbose mode ")
                        
    parser.add_argument('--extra', nargs='*',metavar='PART',help="Add paths to extra yarrrml or rml files to be executed within the parser. Use it at you own risk. Usfull for adding custom triples to the graph. ")
    parser.add_argument('--to_console',action="store_false",help=" False to desactivate logging on cmd (stdout) (default: true) ")
    return parser.parse_args()


def is_cache_activated(cache,process):
    return cache and isinstance(cache,list) and (process in cache or "all" in cache)



            
def separate_extra_paths(extra_paths,work_dir):
    yarrml_paths=[]
    rml_paths=[]
    for path in extra_paths:
        extension=None
        
        path = combine_work_folder(work_dir,path)
        
        if os.path.isfile(path):
            extension = os.path.splitext(path)[1]
            
            
        if extension ==".yaml":
            yarrml_paths.append(path)
        if extension ==".ttl":
            rml_paths.append(path)
    return yarrml_paths,rml_paths

def is_onnx_file(path):
    return os.path.isfile(path) and path.lower().endswith('.onnx')

from pathlib import Path

FILE_CACHE_SEPARATOR="-.-"

def rebuild_onnx_path(files,cache_path,input_file_path):
    new_files =[]
    for file in files:
        cache_folder = os.path.join(*os.path.dirname(os.path.relpath(file, start=cache_path)).split(FILE_CACHE_SEPARATOR))
        model_name = os.path.basename(cache_folder)
        onnx_original_path = os.path.join(input_file_path,os.path.dirname(cache_folder),f"{model_name}.onnx").replace("\\","/")
        new_files.append(Path(onnx_original_path).resolve())
    return new_files


def get_onnx_file(directory, recursive=False,name_model=None,cache_activated=False,cache_path=""):
    base = Path(directory).resolve()
    base_cache = Path(cache_path).resolve()
    regex_cache =LOADED_CACHE 

    
    regex_onnx ="*.onnx"
    regex=".onnx"
    
    results=[]
    paths=[]
    
    if name_model ==None:
        name_model=os.path.basename(directory)

    if recursive:
        paths = list(base.rglob(regex_onnx))
    else:
        paths = [f for f in base.iterdir() if f.is_file() and f.suffix.lower() == regex]
        
    if cache_activated and recursive:
        paths_c = list(base_cache.rglob(regex_cache))
        paths_c = rebuild_onnx_path(paths_c,cache_path,directory)
        paths = set(paths) | set(paths_c)
    elif cache_activated:
        paths_c = [f for f in base_cache.iterdir() if f.is_file() and f.suffix.lower() == regex_cache]
        paths_c = rebuild_onnx_path(paths_c,cache_path,directory)
        paths = set(paths) | set(paths_c)
        
    for file in paths:
        relative_folder = file.parent.relative_to(base)
        
        file_stem = file.stem  
        if relative_folder.parts:
            identifier = f"{str(relative_folder).replace(os.sep, FILE_CACHE_SEPARATOR)}{FILE_CACHE_SEPARATOR}{file_stem}"
        else:
            identifier = f"{file_stem}"
        
        results.append((f"{name_model}/{identifier}", file.resolve().as_posix()))


    return results



def validate_path_arg(arg_name):
    '''The `validate_path_arg` function is a decorator in Python that validates a function argument
    representing a file path.
    
    Parameters
    ----------
    arg_name
        The `arg_name` parameter in the `validate_path_arg` function is used to specify the name of the
    argument that should be validated as a path. The decorator created by this function will then check
    if the value passed to the specified argument is a non-empty string and a valid file path.
    
    Returns
    -------
        The `validate_path_arg` function returns a decorator function that can be used to validate a
    specific argument in another function. The decorator function performs checks on the specified
    argument to ensure it is a non-empty string and a valid file path. If the argument fails the
    validation checks, a `ValueError` is raised with an appropriate error message.
    
    '''
    
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            from inspect import signature
            bound_args = signature(func).bind(*args, **kwargs)
            bound_args.apply_defaults()

            path = bound_args.arguments.get(arg_name)
            if not isinstance(path, str) or not path.strip():
                raise ValueError(f"Argument '{arg_name}' must be a non-empty string.")

            try:
                validate_filepath(path)
            except ValidationError as e:
                raise ValueError(f"Invalid path in argument '{arg_name}': {e}")

            return func(*args, **kwargs)
        return wrapper
    return decorator
import signal
import platform
if platform.system().lower() == "windows":
    signals_to_catch = [signal.SIGINT, signal.SIGTERM, signal.SIGBREAK,signal.SIGABRT]
    dump_redirect = "> nul 2>&1"
else:
    signals_to_catch = [
            signal.SIGINT,     # Ctrl+C
            signal.SIGTERM,    # kill <pid> or default docker stop
            signal.SIGQUIT,    # Ctrl+\
            signal.SIGHUP,     # Terminal closed / systemd reload
            signal.SIGABRT,    # abort()
            signal.SIGPIPE     # Broken pipe
        ]
    dump_redirect = "> /dev/null 2>&1"



class ONNX2RDFParser():
    def cleanup(self):
        if hasattr(self,"__console__"):
            cleanup_handlers(self.__console__)
        if hasattr(self,"logger"):
            cleanup_handlers(self.logger)
    def __del__(self):
        self.cleanup()
        
        
    def __init__(self):
        self.target_path="rdfs"
        self.rdf_format:RDF_formats=RDF_formats.NQUADS
        self.log_folder="logs"
        self.log_extra=False
        self.log_persistant=False
        self.error_folder="errors"
        self.cache=[]
        self.debug=False
        self.tmp_folder="tmp"
        self.work_folder=os.getcwd()
        self.stop_parsing=False
        self.verbose=False
        self.to_console=True
        self.__console__ = setup_logging([],to_console=True,log_id="console")
        self._stop=False
        self._original_handler={}
        self._max_ram="2048m"
        
        download_rmlmapper_jar()
        
        
    


    def __signal_handler__(self,signum, frame):
        self.stop()
    def __signal_handler_stop__(self,signum, frame):
        self.stop()
        raise RuntimeError("Stop Program")
    
    def stop(self):
        
        self._stop=True
        
    def __check_is_stoped__(self):
        if self._stop:
            raise RuntimeError("User Stopped Parsing")
        
    def get_parser_heap(self):
        return self._max_ram
    
    def set_parser_heap(self,heap_size:str):
        self._max_ram=heap_size
    
    def set_rdf_format(self,format:Union[RDF_formats, str]):
        
        extension = get_extension(format)
        if extension==None:
            raise ValueError(f"invalid \"rdf_format\" ({format}) only {FORMATS_EXTENSIONS.keys()} are valid")
        if isinstance(format,str):
            self.rdf_format = RDF_formats._value2member_map_[format]
        else:
            self.rdf_format=format
    
    
    def get_rdf_format(self):
        return self.rdf_format
    
    @validate_path_arg('path')
    def set_log_folder(self,path:str):
        self.log_folder=path
    
    def get_log_folder(self):
        return combine_work_folder(self.work_folder,self.log_folder)
    @validate_path_arg('path')
    def set_tmp_folder(self,path:str):
        self.tmp_folder=path
        
    def get_tmp_folder(self):
        return combine_work_folder(self.work_folder,self.tmp_folder)
    @validate_path_arg('path')
    def set_target_path(self,path:str):
        self.target_path=path
    def get_target_path(self):
        return combine_work_folder(self.work_folder,self.target_path)
    @validate_path_arg('path')
    def set_work_folder(self,path:str):
        path = combine_work_folder(os.getcwd(),path)
        self.work_folder=path
        os.makedirs(path,exist_ok=True)
        
    def get_work_folder(self):
        return self.work_folder
    @validate_path_arg('path')
    def set_error_folder(self,path:str):
        self.error_folder=path
        
    def get_error_folder(self):
        return combine_work_folder(self.work_folder,self.error_folder)
    
    def set_log_extra_files(self,option:bool):
        self.log_extra=option
    def set_log_persistant(self,option:bool):
        self.log_persistant=option
    def set_cache_options(self,parts:list[str]):
        
        if isinstance(parts,str):
            parts=[parts]
        
        self.check_cache_param(parts)
        
        self.cache=parts
    def set_debug(self,option:bool):
        self.debug=option
    def set_stop_parsing(self,option:bool):
        self.stop_parsing=option
    def set_verbose(self,option:bool):
        self.verbose=option
    def set_to_console(self,option:bool):
        self.to_console=option
        
    def get_log_extra_files(self):
        return self.log_extra
    def get_log_persistant(self):
        return self.log_persistant
    def get_cache_options(self):
        return self.cache
    def get_debug(self):
        return self.debug
    def get_stop_parsing(self):
        return self.stop_parsing
    def get_verbose(self):
        return self.verbose
    def get_to_console(self):
        return self.to_console
    
    
    
    def check_cache_param(self,cache = None):
        if cache == None or not isinstance(cache,list):
            return False
        
            
        if cache and len(cache) == 0:
            cache_parts = ['all']
        else:
            invalid = [p for p in cache if p not in VALID_CACHE_PARTS]
            if invalid: 
                self.__console__.warning(f"Invalid cache part(s): [{', '.join(invalid)}] Ignoring Value \n Valid Values:{VALID_CACHE_PARTS}")
            cache_parts = [p for p in cache if p in VALID_CACHE_PARTS]    
        if len(cache_parts)<=0:
            cache=False  
        else:
            cache = cache_parts
        return cache    
    
    
    def run_command(self,command):
        proc = subprocess.Popen(command,shell=True)
        try:
            while proc.poll() is None:
                time.sleep(0.5)  
                self.__check_is_stoped__()
        except Exception:
            proc.terminate()
            raise
        
        out, err = proc.communicate()

        if proc.returncode != 0:
            raise subprocess.CalledProcessError(proc.returncode, command, output=out, stderr=err)
        
    def __set_with_args__(self,args):
        self.set_target_path(args.target_path)
        self.set_rdf_format(args.rdf_format)
        self.set_log_persistant(args.log_persistant)
        self.set_log_extra(args.log_extra)
        if args.cache==None:
            cache=[]
        else:
            cache=args.cache
        self.set_cache_options(cache)
        self.set_log_folder(args.log_folder)
        self.set_error_folder(args.error_folder)
        self.set_debug(args.debug)
        self.set_tmp_folder(args.tmp_folder)
        self.set_verbose(args.verbose)
        self.set_stop_parsing(args.stop_parsing)
        if args.work_folder and args.work_folder != "":
            self.set_work_folder(args.work_folder)
        self.set_to_console(args.to_console)
        
    @staticmethod
    def __build_setting_report__(run_info):
        #TODO: print variables chosen for parse (example logging.info(cache setting))
        pass 
    @staticmethod    
    def __build_result_report__(errors_found,warnings_caught,run_info:dict,error_type:ErrorsONNX2RDF=ErrorsONNX2RDF.NONE_ERROR):
        run_info_c=run_info.copy()
        run_info_c["global_elapsed_time"]=time.time()-run_info_c["start_time"]
        
        result={"errors_found":errors_found,"warnings_caught":warnings_caught,"run_info":run_info_c,"error_type":error_type}

        return result
    @staticmethod
    def __acumulate_times_report__(key,report:dict,variable):
        acum_variable=variable
        if key in report.keys():
            acum_variable = acum_variable + report[key]
        else:
            acum_variable=-1
        return acum_variable 
    @staticmethod   
    def __build_global_results_report(result_entries,stoped,uris):
        result={"result_entries":result_entries}
        global_found_erros =False
        global_found_warnings=False

        
        load_elapsed_time=0
        preprocess_elapsed_time=0
        yarrr2rml_elapsed_time=0
        rml_parsing_elapsed_time=0
        global_elapsed_time=0
        errors=[] 
        for result_entry in result_entries:
            if "errors_found" in result_entry and result_entry["errors_found"]:
                global_found_erros=True
            if "warnings_caught" in result_entry and result_entry["warnings_caught"]:
                global_found_warnings=True
            load_elapsed_time = ONNX2RDFParser.__acumulate_times_report__("load_elapsed_time",result_entry["run_info"],load_elapsed_time)
            preprocess_elapsed_time = ONNX2RDFParser.__acumulate_times_report__("preprocess_elapsed_time",result_entry["run_info"],preprocess_elapsed_time)
            yarrr2rml_elapsed_time = ONNX2RDFParser.__acumulate_times_report__("yarrr2rml_elapsed_time",result_entry["run_info"],yarrr2rml_elapsed_time) 
            rml_parsing_elapsed_time = ONNX2RDFParser.__acumulate_times_report__("rml_parsing_elapsed_time",result_entry["run_info"],rml_parsing_elapsed_time) 
            global_elapsed_time = ONNX2RDFParser.__acumulate_times_report__("global_elapsed_time",result_entry["run_info"],global_elapsed_time)     
            errors.append(result_entry["error_type"])
        
        result["error_types"]=errors
        result["number_models"]=len(result_entries)
        
        result["load_elapsed_time"]=load_elapsed_time
        result["preprocess_elapsed_time"]=preprocess_elapsed_time 
        result["yarrr2rml_elapsed_time"]=yarrr2rml_elapsed_time
        result["rml_parsing_elapsed_time"]=rml_parsing_elapsed_time
        result["global_elapsed_time"]=global_elapsed_time
               
        result["errors_found"]=global_found_erros
        result["warnings_caught"]=global_found_warnings
        result["stopped"]=stoped
        result["model_uris"]=uris
        return result
    
    def transform_yarrrml_rml(self,mapping_path,extra_mappings,cache=False):
        #TODO add rml file to logging folder if log extra
        
        mapping_dir = os.path.dirname(mapping_path)
        mapping_name = os.path.splitext(os.path.basename(mapping_path))[0]
        ttl_mapping_path = os.path.join(mapping_dir,mapping_name+".ttl")
        cache_activated=False
        exists_file = os.path.exists(ttl_mapping_path)
        if cache and exists_file:
            cache_activated=True
        extra_mapping_str=""
        if len(extra_mappings)>0:
            extra_mapping_str = f"-i {' -i '.join(extra_mappings)}"
        
        if not cache or not exists_file:
            comando = f"yarrrml-parser -i {mapping_path} {extra_mapping_str} -o {ttl_mapping_path} {dump_redirect} "
            self.run_command(comando)
        
        return ttl_mapping_path,cache_activated
    
    
    
    def yarrml2_rdf_pipeline(self,yarrml_path,file_name="result",output_folder="result_extra"):
      
     
        script_path = os.path.dirname(os.path.abspath(__file__))
        rml_parser = os.path.join(script_path,MAPPER_NAME)
        extension = get_extension(self.get_rdf_format())    
        rdf_file_name = file_name+extension
        ttl_file_name = f"{os.path.splitext(os.path.basename(yarrml_path))[0]}.ttl"
        ttl_mapping_path = combine_work_folder(self.work_folder,os.path.dirname(yarrml_path),ttl_file_name)
        output_path = combine_work_folder(self.work_folder,output_folder,rdf_file_name)
        max_ram = self._max_ram

        os.makedirs(os.path.dirname(output_path), exist_ok=True)
        os.makedirs(os.path.dirname(ttl_mapping_path), exist_ok=True)
        
        
        
        
        comando = f"yarrrml-parser -i {yarrml_path} -o {ttl_mapping_path} {dump_redirect} "
        self.run_command(comando)
      
        comando = f"java -Xmx{max_ram} -jar {rml_parser} -m {ttl_mapping_path} -o {output_path} -s {self.get_rdf_format().value}"
        self.run_command(comando)
        return output_path
        
    def rml_parsing(self,rml_parser,work_dir,mapping_path,extra_mappings,output_folder="rdf",model_name_path="model",output_format:RDF_formats=RDF_formats.NQUADS,verbose=False):

        
        extension = get_extension(output_format)
        
        
        rdf_file_name = model_name_path+extension
        
        extra_mapping_str=""
        if len(extra_mappings)>0:
            extra_mapping_str = f"-m  {' -m '.join(extra_mappings)}"
        
        
        output_path = combine_work_folder(work_dir,output_folder,rdf_file_name)

        os.makedirs(os.path.dirname(output_path), exist_ok=True)
        
        
        max_ram = self._max_ram
        
        comando = f"java -Xmx{max_ram} -jar {rml_parser} -m {mapping_path} {extra_mapping_str} -o {output_path} -s {output_format.value}"
        if verbose:
            comando = comando +" -v"
        self.run_command(comando)
        return output_path
    
    
    
        
    def __pipeline_run__(self,input_file_path,model_name_path=None,base_resource_url=None,id_process="",extra_files=[]):
        
        
        time_stamp = datetime.now()
        start_time= time.time()
        script_path = os.path.dirname(os.path.abspath(__file__))
        rml_parser = os.path.join(script_path,MAPPER_NAME)
        
        to_console=self.to_console
        
        if base_resource_url==None:
            base_resource_url=DEFAULT_RESOURCE_BASE_URL

        
        warning_caught =False
        
        work_folder=self.work_folder
        tmp_folder=self.tmp_folder
        rdf_format = self.rdf_format
        target_path = self.target_path
        log_folder = self.log_folder
        log_extra = self.log_extra
        log_persistant = self.log_persistant
        error_folder = self.error_folder
        cache = self.cache
        debug=self.debug
        verbose=self.verbose
        stop_parsing=self.stop_parsing
        
        
        run_info=dict()
        
        run_info["input_file_path"]=input_file_path
        run_info["work_folder"]=work_folder
        run_info["log_folder"]=log_folder
        run_info["error_folder"]=error_folder
        run_info["tmp_folder"]=tmp_folder
        run_info["datetime"]=time_stamp
        run_info["start_time"]=start_time
        run_info["target_path"]=target_path
        run_info["rdf_format"]=str(rdf_format)
        run_info["id_process"]=id_process
        run_info["verbose"]=verbose!=None
        run_info["cache"]=cache
        run_info["log_extra"]=log_extra
        run_info["extra_paths"]=extra_files
        run_info["model_name_path"]=model_name_path
        
        if extra_files==None:
            extra_files=[] 
        extra_yarrml_paths,extra_rml_paths=  separate_extra_paths(extra_files,work_folder)
        
        
        error_folder_specific, error_log_file = setup_error_file(error_folder,work_folder,model_name_path)
        
        
        if id_process!="":
            log_file_name = f"logs_{id_process}.log"    
            default_log_folder = "last_execution"+"_"+str(id_process)
        else:
            default_log_folder = "last_execution"
            log_file_name = "logs.log"
        
        default_execution_folder = os.path.join(log_folder,default_log_folder) 
        log_dirs=[default_execution_folder]
        if os.path.exists(default_execution_folder):
            remove_files(default_execution_folder)
        
        log_paths=[]
        if log_persistant:
            time_str = time_stamp.strftime(model_name_path+"_%d_%m_%Y_%Hh_%Mmin")
            time_folder = os.path.join(log_folder,time_str)
            log_dirs.append(time_folder)
        log_paths = get_log_paths(log_dirs,log_file_name,work_folder)
          
        self.logger =setup_logging(log_paths,to_console=to_console,error_file=error_log_file,id_process=id_process)
        logger=self.logger
        if not log_extra:   
            log_dirs=[] 
        if log_persistant and log_extra:
            log_dirs=time_folder
        
            
        
            
        
            
        cache = self.check_cache_param(cache)
        if cache or isinstance(cache,list) and len(cache)>0:
            logger.info(f"Cache settings: {cache}")
        else:
            logger.info("Cache settings: Disable")
            cache=[]
            
            
        onnx_tmp_path = os.path.join(work_folder,os.path.join(tmp_folder,model_name_path),LOADED_CACHE)
            
        if not os.path.exists(input_file_path) and not is_cache_activated(cache,"load-model"):
            logger.error(f"Error: The model at path {input_file_path} does not exist and cache is not activated.")
            process_errors(ErrorsONNX2RDF.LOADING_MODEL,specific_error_folder=error_folder_specific,run_info=run_info,logger=logger,stopped=self._stop)
            cleanup_handlers(self.logger)
            
            return self.__build_result_report__(True,warning_caught,run_info,error_type=ErrorsONNX2RDF.FILE_NOT_PRESENT)
        elif not os.path.exists(input_file_path) and not os.path.exists(onnx_tmp_path):
            logger.error(f"Error: The model at path {input_file_path} does not exist neither its cache file.")
            process_errors(ErrorsONNX2RDF.LOADING_MODEL,specific_error_folder=error_folder_specific,run_info=run_info,logger=logger,stopped=self._stop)
            cleanup_handlers(self.logger)
            return self.__build_result_report__(True,warning_caught,run_info,error_type=ErrorsONNX2RDF.CACHE_FILE_NOT_PRESENT)

        cache_execute_load=False

        logger.info(f"ONNX loading/cleaning started | Model:({model_name_path})")
        

        
        start_load=time.time()
        
        
        try:
            self.__check_is_stoped__()
            json_data,cache_execute_load= execute_clean_model(input_file_path,
                                    tmp_path=onnx_tmp_path,
                                    relative_dir=work_folder,
                                    log_dirs=log_dirs,
                                    cache=is_cache_activated(cache,"load-model"))
        except Exception:
            if not self._stop:
                logger.error(f"Loading file of Model:({model_name_path}) got unexcpeted error\n {traceback.format_exc()}")
            process_errors(ErrorsONNX2RDF.LOADING_MODEL,specific_error_folder=error_folder_specific,run_info=run_info,logger=logger,stopped=self._stop)
            cleanup_handlers(logger)   
            return self.__build_result_report__(True,warning_caught,run_info,ErrorsONNX2RDF.LOADING_MODEL)
        run_info["load_elapsed_time"]=time.time()-start_load
        
        
        
        if cache_execute_load:
            logger.info(f"Cache for [load-model] was activated: File '{onnx_tmp_path}' was used")
        
        run_info["onnx_tmp_path"] = onnx_tmp_path
        start_preprocess_time=time.time()
            
        cache_execute_preprocess=False
        
        logger.info(f"ONNX preprocess started | Model:({model_name_path})")
        
        with warnings.catch_warnings(record=True) as w:
            error_found=False
            error_messege=None
            try:
                self.__check_is_stoped__()
                
                self.set_multiple_singal(signals_to_catch,self.__signal_handler_stop__)
                pre_tmp_path,cache_execute_preprocess = execute_pre_process(json_data,
                                        relative_dir=work_folder,
                                        log_dirs=log_dirs,
                                        tmp_dir=os.path.join(tmp_folder,model_name_path),
                                        cache=is_cache_activated(cache,"pre-process"))
                
                self.set_multiple_singal(signals_to_catch,self.__signal_handler__)
            except Exception:
                error_found=True
                error_messege=traceback.format_exc()
            except (BaseException):
                #Keyboard Interrupt or similars
                self._stop=True
            
            
                
            if w and not self._stop:
                
                w_filtered = []
                current_thread = threading.current_thread().name
                
                for warning in w:
                    
                    message_str = str(warning.message)
                    match = re.match(r"^\[([^\]]+)\]\s*(.*)", message_str)
                    thread_name=None
                    if match:
                        thread_name = match.group(1)
                        message = match.group(2)
                    else:
                        message=message_str
                    
                    if thread_name and thread_name ==current_thread:
                        w_filtered.append(message)
                            
                logger.warning(f"ONNX preprocess finish - {len(w_filtered)} warning(s) caught | Model:({model_name_path})")
                warning_caught=True
                
                for warning_messege in w_filtered:

                    logger.warning(f" Preproces Warning | Model:({model_name_path}) {warning_messege}")
            elif not self._stop:
                logger.info(f"ONNX preprocess finish - No warnings were raised | Model:({model_name_path}).")
            if error_found:
                if not self._stop:
                    logger.error(f"Preprocesing file of Model:({model_name_path}) got unexcpeted error\n {error_messege}")
                process_errors(ErrorsONNX2RDF.PREPROCESING_MODEL,specific_error_folder=error_folder_specific,run_info=run_info,logger=logger,stopped=self._stop)
                cleanup_handlers(logger)
                return self.__build_result_report__(True,warning_caught,run_info)
            
        run_info["process_tmp_path"]=pre_tmp_path  
        run_info["preprocess_elapsed_time"]=time.time()-start_preprocess_time   
            
        if cache_execute_preprocess:
            logger.info(f"Cache for [pre-process] was activated: File '{pre_tmp_path}' was used ")
        
        
        cache_activated_copy_edit=False
        
        yarrml_mapping = os.path.join(script_path,"mappings/mapper_yarr.yaml")
        run_info["yarrml_mapping"]=yarrml_mapping
        start_yarrr2rml_time=time.time()
        
        # YARRM_RML PROCESS
        logger.info(f"YARR2_RML started | Model:({model_name_path})")
        
        try:
            self.__check_is_stoped__()
            uri_name = model_name_path.replace(FILE_CACHE_SEPARATOR,".")
            model_uri = f"{base_resource_url}{uri_name}/"
            
            mapping_path,cache_activated_copy_edit = copy_edit_mappings(yarrml_mapping,work_folder,pre_tmp_path,id_process=id_process,
                                                                        log_dirs=log_dirs,
                                                                        cache=is_cache_activated(cache,"yamml2rml"),
                                                                        model_full_uri=model_uri,
                                                                        tmp_dir=os.path.join(tmp_folder,"mappings"))

        except Exception:
            if not self._stop:
                logger.error(f"Modifying the yarrml before transforming of Model:({model_name_path}) got unexcpeted error\n {traceback.format_exc()}")
            process_errors(ErrorsONNX2RDF.YAMML2RML_COPY,specific_error_folder=error_folder_specific,run_info=run_info,logger=logger,stopped=self._stop)
            cleanup_handlers(logger)
            return self.__build_result_report__(True,warning_caught,run_info)
        
        run_info["mapping_path"]=mapping_path 
        cache_activated_yarr2rml=False
        try:
            mapping_rml_path,cache_activated_yarr2rml = self.transform_yarrrml_rml(mapping_path,extra_yarrml_paths,cache=is_cache_activated(cache,"yamml2rml"))
        except Exception:
            
            logger.error(f"Transforming yarrml to rml of Model:({model_name_path}) got unexcpeted error\n {traceback.format_exc()}")
            process_errors(ErrorsONNX2RDF.YAMML2RML,specific_error_folder=error_folder_specific,run_info=run_info,logger=logger,stopped=self._stop)
            cleanup_handlers(logger)
            return self.__build_result_report__(True,warning_caught,run_info)
        run_info["mapping_rml_path"]=mapping_rml_path 
        
        run_info["yarrr2rml_elapsed_time"]=time.time()-start_yarrr2rml_time    
         
            
        files_cache_yaml2rml=[]    
        if cache_activated_copy_edit:
            files_cache_yaml2rml=[mapping_path]
        if cache_activated_yarr2rml:
            files_cache_yaml2rml.append(mapping_rml_path) 
        if len(files_cache_yaml2rml)==1:
            logger.info(f"Cache for [yamml2rml] was activated: File '{files_cache_yaml2rml[0]}' was used ")
        if len(files_cache_yaml2rml)>1:
            logger.info(f"Cache for [yamml2rml] was activated: Files {files_cache_yaml2rml} were used ") 
        
        
        
        start_rml_parsing_time=time.time()
        # MAPPING PROCESS
        if not stop_parsing:
            logger.info(f"RML Parsing started | Model:({model_name_path})")
            try:
                self.__check_is_stoped__()
                output_path = self.rml_parsing(rml_parser,work_folder,mapping_rml_path,extra_rml_paths,target_path,model_name_path,rdf_format,verbose)
                run_info["output_path_rdf"]=output_path
            except Exception:
                if not self._stop:
                    logger.error(f"Mapping rml process of Model:({model_name_path}) got unexcpeted error\n {traceback.format_exc()}")
                process_errors(ErrorsONNX2RDF.MAPPING,specific_error_folder=error_folder_specific,run_info=run_info,logger=logger,stopped=self._stop)
                cleanup_handlers(logger)
                return self.__build_result_report__(True,warning_caught,run_info)
            logger.info(f"Parsing Finished | Model:({model_name_path})")
        else:
            logger.info(f"RML Parsing Skipped --stop-parsing | Model:({model_name_path})")
        run_info["rml_parsing_elapsed_time"]=time.time()-start_rml_parsing_time   

        # Remove temporary files

        
        if not warning_caught and not debug and not is_cache_activated(cache,"pre-process") and os.path.exists(pre_tmp_path):
            try:
                os.remove(pre_tmp_path)
            except Exception:
                pass
     
        if not warning_caught and not debug and not is_cache_activated(cache,"load-model") and os.path.exists(onnx_tmp_path):
            try:    
                os.remove(onnx_tmp_path)
            except Exception:
                pass
        remove_folder_rec(tmp_folder,work_folder,model_name_path)
                
        if not warning_caught and not debug and not is_cache_activated(cache,"yamml2rml"):
            if os.path.exists(mapping_path):
                try:
                    os.remove(mapping_path)
                except Exception:
                    pass
            if os.path.exists(mapping_rml_path):
                try:
                    os.remove(mapping_rml_path)
                except Exception:
                    pass
            mapping_folder = os.path.dirname(mapping_path)
            if os.path.exists(mapping_folder) and is_folder_empty(mapping_folder):
                try:
                    os.rmdir(mapping_folder)
                except Exception:
                    pass
        cleanup_handlers(logger)        
        if not warning_caught and os.path.exists(error_folder_specific):
            remove_files(error_folder_specific)
            remove_folder_rec(error_folder,work_folder,model_name_path)
        
        if warning_caught:
            process_errors(ErrorsONNX2RDF.WARNING_ON_PREPROCESS,specific_error_folder=error_folder_specific,run_info=run_info,logger=logger,stopped=self._stop)
                
        
        return self.__build_result_report__(False,warning_caught,run_info)
    
    def set_multiple_singal(self,signal_types:list[signal.Signals],handle):
        if threading.current_thread() == threading.main_thread():
            for signal_type in signal_types:
                if ONNX2RDFParser.is_signal_editable(signal_type):
                    signal.signal(signal_type,handle)
    def store_original_handlers(self,signal_types:list[signal.Signals]):
        for signal_type in signal_types:
            self._original_handler[signal_type]=signal.getsignal(signal_type)
    @staticmethod       
    def is_signal_editable(signal_type):
        value = None
        if isinstance(signal_type,signal.Signals):
            value = signal_type.value
        if isinstance(signal_type,int):
            value = signal_type
        return value and value!=9 and value!=19
        
    def restore_multiple_singal(self,signal_types:list[signal.Signals]):
        if threading.current_thread() == threading.main_thread():
            for signal_type in signal_types:
                if ONNX2RDFParser.is_signal_editable(signal_type):
                    signal.signal(signal_type,self._original_handler[signal_type]) 


    def parse_file(self,model_path,model_name=None,base_resource_url=None,id_process="",extra_files=[]):
        result_entries=[]
        try:
            
            error_if_no_java()
            
            self.store_original_handlers(signals_to_catch)
            self.set_multiple_singal(signals_to_catch,self.__signal_handler__)
            
            cache = self.check_cache_param(self.cache)
            
            
            work_folder=self.work_folder
            input_file_path = combine_work_folder(work_folder,model_path)
                
            if os.path.isdir(input_file_path):
                if model_name==None:
                    cache_path = os.path.join(work_folder,self.tmp_folder,os.path.basename(input_file_path))
                else:
                    cache_path = os.path.join(work_folder,self.tmp_folder,model_name)
                onnx_files = get_onnx_file(input_file_path,recursive=True,name_model=model_name,cache_activated=is_cache_activated(cache,"load-model"),cache_path=cache_path)
                
                if len(onnx_files)==0 and not is_cache_activated(cache,"load-model"):
                    raise ValueError(f"No .onnx files found at {input_file_path}")
                elif len(onnx_files)==0:
                    raise ValueError(f"No loaded_model.json cache file or .onnx files found at tmp folder at folder or input_folder {cache_path}")
            else:
                
                file_name = os.path.splitext(os.path.basename(input_file_path))[0]
                
                
                if model_name==None:
                    cache_path = os.path.join(work_folder,self.tmp_folder,file_name,"model",LOADED_CACHE)
                else:
                    cache_path = os.path.join(work_folder,self.tmp_folder,model_name,file_name,LOADED_CACHE)
                
                is_onnx =not is_onnx_file(input_file_path)
                if is_onnx and not is_cache_activated(cache,"load-model"):
                    raise ValueError(f" model_path ({input_file_path}) is not a .onnx file")
                elif is_onnx and not os.path.exists(cache_path):
                    raise ValueError(f" model_path ({input_file_path}) is not a .onnx file and there is not cache file")
                
                
                if model_name==None:
                    model_name_path=os.path.join(file_name,"model")
                else:
                    model_name_path=os.path.join(model_name,file_name)
                onnx_files = [(model_name_path,input_file_path)]

            
            if base_resource_url==None:
                base_resource_url=DEFAULT_RESOURCE_BASE_URL
            result_entries = []    
                
            try:
                for onnx_file in onnx_files:
                    result_entry = self.__pipeline_run__(input_file_path=onnx_file[1],
                                        model_name_path=onnx_file[0],
                                        id_process=id_process,
                                        base_resource_url=base_resource_url,
                                        extra_files=extra_files)
                    
                    result_entries.append(result_entry)
                    if self._stop:
                        self.logger.error("ONXX2RDF Parser Stopped Cleanly (It can return None)")
                        break
            except Exception:
                if hasattr(self,"logger"):
                    cleanup_handlers(self.logger)
                raise
            uris = __get_all_uri_models__(onnx_files)    
                
            cleanup_handlers(self.logger)
            
            if result_entries==[]:
                return {"stopped":True}
            return self.__build_global_results_report(result_entries,self._stop,uris)
        except (BaseException):
            
            
            self.restore_multiple_singal(signals_to_catch)
            if self._stop:
                self.__console__.error("ONXX2RDF Parser Stopped Abruptly")
            else:
                self.__console__.error(f"ONXX2RDF Parser got unexcpeted error\n {traceback.format_exc()}")
            if hasattr(self,"logger"):
                cleanup_handlers(self.logger)
            if not result_entries:
                return {"stopped":True}
            
            return self.__build_global_results_report(result_entries,True)
            
            
            
def __get_all_uri_models__(onnx_files):
    uris=[]
    for onnx_file in onnx_files:
        path = onnx_file[0].replace(FILE_CACHE_SEPARATOR,".")
        uris.append(path)   
    return uris
        


def is_folder_empty(path):
    return not any(os.listdir(path))




    
  


import shutil



def process_errors(error_type:ErrorsONNX2RDF,specific_error_folder,run_info=None,logger=None,stopped=False):
    
    try:
        meta_file_path = os.path.join(specific_error_folder,"error_data.json")
        
        meta_data ={"name":run_info["model_name_path"],"error_type":error_type.name,"datetime":run_info["datetime"].strftime("%H_%M_%d_%m_%Y"),"paths":
            {"work_folder":run_info["work_folder"],"log_folder":run_info["log_folder"],"temporary_folder":run_info["tmp_folder"],
            "error_folder":run_info["error_folder"],"rdf_path":run_info["target_path"]},"verbose":run_info["verbose"],"cache":run_info["cache"],
            "extra_paths":run_info["extra_paths"]}
        
        
        meta_data["loading_model"]={"model_path":run_info["input_file_path"]}
        
        
        if error_type.value>1:
            new_path = os.path.join(specific_error_folder,LOADED_CACHE)
            file_info = {"original_path":run_info["onnx_tmp_path"],"report_path":new_path}
            meta_data["loading_model"]["result_loading"] = file_info
            shutil.copy2(run_info["onnx_tmp_path"], new_path)
            meta_data["pre_process"]={"loaded_model_path":file_info}
            
        if error_type.value>2:
            
            new_path = os.path.join(specific_error_folder,"preprocess_model.json")
            file_info = {"original_path":run_info["process_tmp_path"],"report_path":new_path}
            meta_data["pre_process"]["result_preprocess"] = file_info
            shutil.copy2(run_info["process_tmp_path"], new_path)
            
            new_path = os.path.join(specific_error_folder,"mapping.yamml")
            file_info_yamml = {"original_path":run_info["yarrml_mapping"],"report_path":new_path}
            shutil.copy2(run_info["yarrml_mapping"], new_path)
            meta_data["copy_mapping"]={"preproces_model_path":run_info["process_tmp_path"],"id_process":run_info["id_process"],
                                    "yarrml_mapping":file_info_yamml}

            
        if error_type.value>3: 
            
            
            new_path = os.path.join(specific_error_folder,"modified_mapping.yamml")
            file_info = {"original_path":run_info["mapping_path"],"report_path":new_path}
            meta_data["copy_mapping"]["result_edit_mapping"] = file_info
            shutil.copy2(run_info["mapping_path"], new_path)
            
            meta_data["yamml2rml"]={"modified_mapping_path":file_info}

            
        if error_type.value>4:
            new_path = os.path.join(specific_error_folder,"modified_mapping.rml")
            file_info = {"original_path":run_info["mapping_rml_path"],"report_path":new_path}
            meta_data["yamml2rml"]["result_rml"] = file_info
            
            file_info_json = {"original_path":run_info["process_tmp_path"],"report_path":new_path}
            shutil.copy2(run_info["mapping_rml_path"], new_path)
            
            meta_data["rml_mapping"]={"rml_mapping":file_info,"preprocess_data":file_info_json,"rdf_format":run_info["rdf_format"]}
        
        if error_type.value>5:
            file_info = {"original_path":run_info["output_path_rdf"]}
            meta_data["rml_mapping"]["result_mapping"] = file_info
            
        if stopped:
            if logger:
                logger.info("Error by Stopped Program")
            meta_data["stopped"]=True
          
        save_json(meta_data,meta_file_path)
    except Exception:
        logger.error(f"Unexcpected Error trying to document and error ,{traceback.format_exc()}")
        
        



def call_main_with_args():
    # Obtener los argumentos desde la línea de comandos
    args = parse_args()
    
    parser  = ONNX2RDFParser()
    
    parser.__set_with_args__(args)
    
    parser.parse_file(args.model_path,args.model_name_path,args.base_url,args.id_process,args.extra)
    


if __name__ == "__main__":
    
    # Ejecutar la función call_main_with_args solo cuando se ejecuta como script
    call_main_with_args()



def error_if_no_java():
    try:
        subprocess.run(["java", "-version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
    except Exception:
        raise RuntimeError("⚠️ WARNING: Java (OpenJDK) not found. This package may not work correctly without Java.")